"""
deprecated! This class has been copied for better retro-compatibility with old projects(e.g. Brutos)
"""

from ... import _hub as inj


pygame = inj.pygame


class Trigger:
    class_ready_to_draw = False
    RED = None
    small_font = None
    med_font = None

    def __init__(self, pos_topleft, size):
        self.pos = pos_topleft
        self.size = size
        self.active = True
        self.visibility = True
        self.label = None
        self.spr_label = None
        self.RED = self.drawing_color = pygame.Color('red')
        self.small_font = pygame.font.Font(None, 15)
        self.med_font = pygame.font.Font(None, 28)

    def set_active(self, val):
        self.active = val

    def is_active(self):
        return self.active

    def set_visible(self, val):
        assert (isinstance(val, bool))
        self.visibility = val

    def is_visible(self):
        return self.visibility

    def getSize(self):
        return self.size

    def setSize(self, new_size):
        self.size = new_size

    def setPos(self, new_pos):
        self.pos = new_pos

    def getPos(self):
        return self.pos

    def getLabel(self):
        return self.label

    def getRect(self):
        return [self.pos, self.size]

    def getCollisionBox(self):
        repr_cb = [self.pos, self.size]

        pt_topleft = repr_cb[0]

        taille_util = repr_cb[1]
        pt_botright = list(pt_topleft)
        pt_botright[0] += taille_util[0]
        pt_botright[1] += taille_util[1]

        res = list()
        res.extend(pt_topleft)
        res.extend(pt_botright)
        return res

    def setCollisionBox(self, new_cb_borders):
        # --- conversion
        taille = (
            new_cb_borders[2] - new_cb_borders[0],
            new_cb_borders[3] - new_cb_borders[1])
        pt_topleft = (new_cb_borders[0], new_cb_borders[1])
        self.setPos(pt_topleft)
        self.setSize(taille)

    def contains(self, point2d):
        if not self.active:
            return False
        tx, ty = point2d
        xbinf, ybinf, xbsup, ybsup = self.getCollisionBox()
        return (xbinf < tx < xbsup) and (ybinf < ty < ybsup)

    def setLabel(self, nouveau_label=None):
        self.label = nouveau_label
        if nouveau_label is None:
            self.spr_label = None
        else:
            self.spr_label = self.med_font.render(nouveau_label, True, self.drawing_color)

    def draw(self, screen, forced_color=None):
        if not self.visibility:  # invisible trigger
            return

        if forced_color is not None:
            self.drawing_color = forced_color
            self.setLabel(self.label)

        c1, c2, t, tt = self.getCollisionBox()
        c3 = t - c1
        c4 = tt - c2
        rect = ((c1, c2), (c3, c4))
        pygame.draw.rect(screen, self.drawing_color, rect, 1)

        if self.label is not None:
            # --- le but est dafficher ce txt au centre de la collision box
            pt_centrage = [
                (t - c1) // 2,
                (tt - c2) // 2,
            ]
            pt_centrage[0] += c1
            pt_centrage[1] += c2

            img = self.spr_label
            tmp_pt = (pt_centrage[0] - img.get_width() // 2, pt_centrage[1] - img.get_height() // 2)
            screen.blit(self.spr_label, tmp_pt)

        if forced_color is not None:
            self.drawing_color = self.RED
            self.setLabel(self.label)



from ... import _hub

pygame = _hub.pygame


def clip(surf, x, y, x_size, y_size):
    # WTF bro? Part of original code commented:
    # handle_surf = surf.copy()
    clip_r = pygame.Rect(x, y, x_size, y_size)
    # handle_surf.set_clip(clipR)
    image = surf.subsurface(clip_r)  # handle_surf.get_clip())

    # pr port ca vers Web Ctx
    # ya donc que 2 operations a emuler convenablement: .subsurface & .copy
    return image.copy()


def swap_color(img, old_c, new_c):
    img.set_colorkey(old_c)
    surf = img.copy()
    surf.fill(new_c)
    surf.blit(img, (0, 0))
    return surf


def load_font_img(path, font_color):
    fg_color = (255, 0, 0)
    bg_color = (0, 0, 0)
    font_img = pygame.image.load(path).convert()

    # N.B. theres a BUG when interactin with KataSDK, thats why I have
    # commetend the line below.
    # Hence ktg-webapp does not crash but it cannot swap color.
    # TODO: fix the bug Asap!

    # font_img = swap_color(font_img, fg_color, font_color)
    last_x = 0
    letters = []
    letter_spacing = []
    for x in range(font_img.get_width()):
        colorinfos = font_img.get_at((x, 0))
        if colorinfos[0] == 127:
            tmpw = x - last_x
            tmph = font_img.get_height()
            letters.append([last_x, 0, tmpw, tmph])
            # letters.append(
            #     clip(font_img, last_x, 0, tmpw, tmph)
            # )
            letter_spacing.append(tmpw)
            last_x = x + 1
        x += 1
    #for letter in letters:
    #    letter.set_colorkey(bg_color)
    card_l = len(letters)
    print('*debug ImgBasedText*')
    print(f' ... source={path}')
    print(f' ... num of letters={card_l}')
    print(f' ... letter_spacings {letter_spacing}')
    return letters, letter_spacing, font_img.get_height()


class ImgBasedFont:
    NO_CK_MODE = False  # this is a hotfix for the webctx where on needs to not use colorkey val

    def __init__(self, path, color):
        self._spr_sheet_like = pygame.image.load(path)

        self.letters_rects, self.letter_spacing, self.line_height = load_font_img(path, color)
        self.font_order = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R',
                           'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                           'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '.', '-',
                           ',', ':', '+', '\'', '!', '?', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '(', ')',
                           '/', '_', '=', '\\', '[', ']', '*', '"', '<', '>', ';', '#', '$', '%', '@', '{', '}', '`',
                           '~']
        self.space_width = self.letter_spacing[self.font_order.index('_')]
        self.base_spacing = 1
        self.line_spacing = 2

    def get_linesize(self):  # stick to pygame interface
        return 11  # TODO implement properly plz! 11 is hardset comes from 11px the PNG height gibson0_font

    def size(self, sample_txt):  # stick to pygame interface
        return self.width(sample_txt), self.get_linesize()

    def width(self, text):
        text_width = 0
        for char in text:
            if char == ' ':
                text_width += self.space_width + self.base_spacing
            else:
                try:
                    idx = self.font_order.index(char)
                except ValueError:
                    # generic char
                    print('cannot comp width for: ', char)
                    idx = self.font_order.index('_')
                text_width += self.letter_spacing[idx] + self.base_spacing

        return text_width

    # --add-on to stick to pygame interface
    def render(self, gtext, antialias, color, bgcolor=None):
        rez = pygame.Surface((self.width(gtext), self.get_linesize()), pygame.SRCALPHA)
        if not self.NO_CK_MODE:
            upink = (255, 0, 255)  # ugly pink
            rez.fill(upink)
        self._xrender(gtext, rez, (0, 0))
        if not self.NO_CK_MODE:
            rez.set_colorkey(upink)
        return rez

    def _xrender(self, text, surf, loc, line_width=0):

        x_offset = 0
        y_offset = 0
        if line_width != 0:
            spaces = []
            x = 0
            for i, char in enumerate(text):
                if char == ' ':
                    spaces.append((x, i))
                    x += self.space_width + self.base_spacing
                else:
                    x += self.letter_spacing[self.font_order.index(char)] + self.base_spacing
            line_offset = 0
            for i, space in enumerate(spaces):
                if (space[0] - line_offset) > line_width:
                    line_offset += spaces[i - 1][0] - line_offset
                    if i != 0:
                        text = text[:spaces[i - 1][1]] + '\n' + text[spaces[i - 1][1] + 1:]

        escaped_chars = ['\n', ' ', '\r']
        for char in text:
            if char not in escaped_chars:
                #surf.blit(self.letters[self.font_order.index(char)], (loc[0] + x_offset, loc[1] + y_offset))
                destpos = (loc[0] + x_offset, loc[1] + y_offset)
                surf.blit(self._spr_sheet_like, destpos, area=self.letters_rects[self.font_order.index(char)] )
                x_offset += self.letter_spacing[self.font_order.index(char)] + self.base_spacing
                continue
            if char == ' ':
                x_offset += self.space_width + self.base_spacing
            elif char == '\n':
                y_offset += self.line_spacing + self.line_height
                x_offset = 0



import kgame_bundle.defs_mco.glvars as glvars
from coremon_main.events import EventReceiver, EngineEvTypes, EventManager, CgmEvent
from katagames_sdk.capsule.networking.EnvoiAsynchrone2 import EnvoiAsynchrone2
import katagames_sdk.capsule.engine_ground.conf_eng as cgmconf


if cgmconf.runs_in_web():
    from browser import ajax


class AsyncMessagingCtrl(EventReceiver):
    mem_timestamp = None

    def __init__(self):
        super().__init__(True)  # is sticky? YES

        # threads avec envoi / attente en cours
        self.set_of_threads = set()
        self._is_web_mode = cgmconf.runs_in_web()

        if EnvoiAsynchrone2.PORT is None:  # classe non initialisée
            EnvoiAsynchrone2.PORT = glvars.server_port
            EnvoiAsynchrone2.HOST_STR = glvars.server_host
            assert glvars.server_host is not None

    def proc_event(self, ev, source):

        if not self._is_web_mode:
            # - traitement via threading (marche bien en LOCAL)
            if ev.type == EngineEvTypes.ASYNCSEND:
                self.on_async_request(ev)
            elif ev.type == EngineEvTypes.LOGICUPDATE:
                self.on_logic_update(ev)

        else:
            # methode avec cb
            if ev.type == EngineEvTypes.ASYNCSEND:
                backupid = ev.num
                def xoncomplete(req):
                    nonlocal backupid
                    if req.status == 200 or req.status == 0:
                        print((f"%d OK asyncsend gives: " % backupid) + req.text)
                        EventManager.instance().post(CgmEvent(EngineEvTypes.ASYNCRECV, num=backupid, msg=req.text))
                    else:
                        print("error ASYNCSEND" + req.text)

                #req = ajax.Ajax()
                #req.bind('complete', on_complete)

                # send a POST request to the url
                print('+AJAX sending+')
                print('http://{}/{}'.format(glvars.server_host, ev.msg))
                print('post method')
                print('num={}'.format(ev.num))
                print('data={}'.format(ev.data))

                if ev.data is not None:
                    ajax.post('http://{}:{}/{}'.format(glvars.server_host, glvars.server_port, ev.msg), data={'data':ev.data}, oncomplete=xoncomplete)
                else:
                    print('[nodata fiels sent]')
                    ajax.post('http://{}:{}/{}'.format(glvars.server_host, glvars.server_port, ev.msg), oncomplete=xoncomplete)

                #req.open('GET', , ev.msg), True)
                #req.set_header('content-type', 'application/x-www-form-urlencoded')
                # send data as a dictionary
                #req.send({'x': 0, 'y': 1}
                #ev.num, ev.msg, ev.data

    def on_async_request(self, ev):
        print('ds on_async_request')


        print('[AsyncMessagingCtrl] processing async request {}'.format(ev))
        id_req = ev.num
        msg_reseau = ev.msg

        thread = EnvoiAsynchrone2(id_req, msg_reseau, ev.data)  # self.__class__.mem_timestamp)

        # fonctionnement à éviter !
        # if ev.cb is not None:
        #    thread.setup_callback(ev.cb)

        self.set_of_threads.add(thread)
        thread.start()

    def on_logic_update(self, event):
        # aucun thread enregistré
        if len(self.set_of_threads) == 0:
            return

        # si tous les threads continuent leur exécution
        signes_activite = [not t.zombie_thread.isSet() for t in self.set_of_threads]
        if all(signes_activite):
            return

        # on sait que l'exécution d'au moins un thread s'est terminée
        new_set = set()
        for t in self.set_of_threads:
            if not t.zombie_thread.isSet():
                new_set.add(t)
                continue
            new_ev = CgmEvent(EngineEvTypes.ASYNCRECV, num=t.id_req, msg=t.response)
            EventManager.instance().post(new_ev)

        self.set_of_threads = new_set  # on "oublie" ainsi les threads zombie

    def __proc_async_command(self, ev):
        raise NotImplementedError  # TODO utilité ?
        # app_const = AppConstants.instance()
        # # if app_const.get_val('DEV_MODE'):  # désactivé sur serv de prod a cause dun gros bug comm réseau!!
        #     #t = EnvoiAsynchrone(ev.hote, ev.ressource_http)
        #     # print('creation & lancement async worker avec msg: {}'.format(ev.ressource_http))
        # t = AsyncWorker(ev.id_commande, ev.ressource_http)
        # self.set_of_threads.add(t)
        # t.start()  # execute .run() sur le Thread en question, dans un fil d'exécution à part



"""
constants only
--------------
in regard to display options,
within KENGI there are only 4 canonical modes for display:

three that are displayed in a 960 x 720 -pixel canvas
'super_retro' (upscaling x3), 'old_school', (upscaling x2), 'hd' (no upscaling)

one that is displayed in a user-defined size canvas and also uses
a pixel-to-pixel mapping just like the 'hd' option
"""
STD_SCR_SIZE = (960, 720)

USEREVENT = 32850  # pygame userevent 2.1.1

FIRST_ENGIN_TYPE = USEREVENT + 1
FIRST_CUSTO_TYPE = FIRST_ENGIN_TYPE + 20  # therefore, 20 is the maximal amount of engine events



"""
+-----------------------------------------------------+
| KENGI [K]atagames [ENGI]ne                          |
| Motto ~ Never slow down the innovation              |
|                                                     |
| Main author: wkta-tom (github.com/wkta)             |
|                                                     |
| an open-source project funded by GAUDIA TECH INC.   |
| https://github.com/gaudiatech/kengi                 |
+-----------------------------------------------------+

 * defines a subset of the pygame library (chosen functions & objects)
  and creates a wrapper around it

 * allows for a swift implementation of two essential design patterns:
   Mediator and Model-View-Contoller

 * provides easy access to data structures useful in game development:
   stacks, matrices, trees, graphs, finite state machines, cellular automata

 * provides algorithms that may be tricky to code but are super-useful:
   A-star, Minimax, a FOV algorithm for a 2D grid based world,

 * is extensible: kengi is capable of receiving custom events and custom
  extensions, for example a custom GUI manager, an isometric engine, or
  antything similar, without requiring any architectural change

 * can run along with the KataSDK but can also be detached, to run independently

 * does not know ANYTHING about whether your code runs in a web browser or not,
  although the engine can be hacked to allow such a possibility

 * incentivizes you, the creator, to write clean readable easy-to-refactor &
  easy-to-reuse code!
"""

from . import _hub as hub
from . import event
from . import pal
from . import struct
# from .compo import vscreen  # important import! As this is called by web_vm
from .compo.vscreen import flip
from .Injector import Injector
from ._BaseGameState import BaseGameState
from .__version__ import ENGI_VERSION
from .compo import vscreen
from .compo import gfx
from .compo.modes import GameModeMger, BaseGameMode
from .foundation import defs
from .ifaces.pygame import PygameIface
from .util import underscore_format, camel_case_format


_active_state = False
_gameticker = None
_multistate_flag = False
_stack_based_ctrl = None
one_plus_init = False
state_stack = None
ver = ENGI_VERSION


class Objectifier:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def _show_ver_infos():
    print(f'KENGI - ver {ENGI_VERSION}, built on top of ')


def is_ready():
    global one_plus_init
    return one_plus_init


def bootstrap_e(info=None):
    """
    ensure the engine is ready to be used

    :param info:
    :return:
    """
    global pygame, one_plus_init, _gameticker
    if one_plus_init:
        return
    del pygame

    def _ensure_pygame(xinfo):
        # replace iface by genuine pygame lib, use this lib from now on
        if isinstance(xinfo, str):
            hub.kengi_inj.register('pygame', xinfo)
        else:
            hub.kengi_inj.set('pygame', xinfo)  # set the module directly, instead of using lazy load

    one_plus_init = True
    _show_ver_infos()
    if info is None:
        info = 'pygame'
    _ensure_pygame(info)

    event.create_manager()
    _gameticker = event.GameTicker()

    # dry import
    vscreen.cached_pygame_mod = hub.pygame


def screen_param(gfx_mode_code, paintev=None, screen_dim=None):
    global _active_state
    if isinstance(gfx_mode_code, int) and -1 < gfx_mode_code <= 3:
        if gfx_mode_code == 0 and screen_dim is None:
            ValueError(f'graphic mode 0 required an extra valid screen_dim argument(provided by user: {screen_dim})')

        # from here, we know that the gfx_mode_code is 100% valid
        conventionw, conventionh = defs.STD_SCR_SIZE
        if gfx_mode_code != 0:
            adhoc_upscaling = gfx_mode_code
            taille_surf_dessin = int(conventionw/gfx_mode_code), int(conventionh/gfx_mode_code)
        else:
            adhoc_upscaling = 1
            taille_surf_dessin = screen_dim
            print(adhoc_upscaling, taille_surf_dessin)
        # ---------------------------------
        #  legacy code, not modified in july22. It's complex but
        # it works so dont modify unless you really know what you're doing ;)
        # ---------------------------------
        if vscreen.stored_upscaling is None:  # stored_upscaling isnt relevant <= webctx
            _active_state = True
            pygame_surf_dessin = hub.pygame.display.set_mode(taille_surf_dessin)
            vscreen.set_virtual_screen(pygame_surf_dessin)
        else:
            print(taille_surf_dessin)
            pygame_surf_dessin = hub.pygame.surface.Surface(taille_surf_dessin)
            vscreen.set_virtual_screen(pygame_surf_dessin)
            vscreen.set_upscaling(adhoc_upscaling)
            if paintev:
                paintev.screen = pygame_surf_dessin
            if _active_state:
                return
            _active_state = True

            if gfx_mode_code:
                pgscreen = hub.pygame.display.set_mode(defs.STD_SCR_SIZE)
            else:
                pgscreen = hub.pygame.display.set_mode(taille_surf_dessin)
            vscreen.set_realpygame_screen(pgscreen)
    else:
        e_msg = f'graphic mode requested({gfx_mode_code}: {type(gfx_mode_code)}) isnt a valid one! Expected type: int'
        raise ValueError(e_msg)


def init(gfc_mode=1, caption=None, maxfps=60, screen_dim=None):
    global _gameticker
    bootstrap_e()

    pygm = hub.pygame
    pygm.init()
    pygm.mixer.init()
    screen_param(gfc_mode, screen_dim=screen_dim)
    if caption is None:
        caption = f'untitled demo, uses KENGI ver {ENGI_VERSION}'
    pygm.display.set_caption(caption)

    _gameticker.maxfps = maxfps


def get_surface():
    global _active_state
    if not is_ready():
        raise Exception('calling kengi.get_surface() while the engine isnt ready! (no previous bootstrap op.)')
    if not _active_state:
        raise Exception('kengi.init has not been called yet')
    return vscreen.screen


def declare_states(gsdefinition, assoc_gscode_cls, mod_glvars=None):
    global _multistate_flag, state_stack, _stack_based_ctrl
    _multistate_flag = True
    state_stack = struct.Stack()
    _stack_based_ctrl = event.StackBasedGameCtrl(
        _gameticker, gsdefinition, mod_glvars, assoc_gscode_cls
    )


def get_game_ctrl():
    global _multistate_flag, _stack_based_ctrl, _gameticker
    if _multistate_flag:
        return _stack_based_ctrl
    else:
        return _gameticker


def get_manager():  # saves some time
    return event.EventManager.instance()


def quit():  # we keep the "quit" name bc of pygame
    global _active_state, _multistate_flag, _stack_based_ctrl

    if not _active_state:
        return

    if _multistate_flag:
        _multistate_flag = False
        _stack_based_ctrl = None
    if hub.ascii.is_ready():
        hub.ascii.reset()
    event.EventManager.instance().hard_reset()
    event.CogObj.reset_class_state()
    vscreen.init2_done = False
    pyg = get_injector()['pygame']
    pyg.mixer.quit()
    pyg.quit()
    _active_state = False


def get_injector():
    return hub.kengi_inj


def plugin_bind(plugin_name, pypath):
    hub.kengi_inj.register(plugin_name, pypath)


def bulk_plugin_bind(darg: dict):
    """
    :param darg: association extension(plug-in) name to a pypath
    :return:
    """
    for pname, ppath in darg.items():
        plugin_bind(ppath, ppath)


# ----------------------------
# Stuff related to lazy import
# ----------------------------
pygame = PygameIface()


def __getattr__(attr_name):
    if not is_ready():
        raise AttributeError(f"kengi cannot lazy load, it hasnt bootstrap yet! (user request: {attr_name})")
    else:
        return getattr(hub, attr_name)



import time

import katagames_sdk.capsule.gui as gui
from katagames_sdk.capsule.bioslike.KataFrameM import KataFrameM
from katagames_sdk.capsule.event import EngineEvTypes, EventReceiver, EventManager, CgmEvent
from katagames_sdk.capsule.pygame_provider import get_module

pygame = get_module()


BIOS_FG_COL_DESC = 'purple'
BIOS_BG_COL_DESC = 'antiquewhite3'


class KataFrameV(EventReceiver):
    """
    affiche écran intermédiaire permettant la SAISIE des IDENT.
    """

    TEXT_INPUT_SIZE = 160
    ERR_MSG_DELAY = 2.5  # sec
    POS_BT_BROWSER = (385/2, 215/2)
    TAQ_XALIGN_BOX = 150
    TAQ_XALIGN_LABELS = 50
    VOFFSET = 24
    POS_BT_CONFIRM = (TAQ_XALIGN_LABELS, 199)
    POS_BT_CANCEL = (TAQ_XALIGN_LABELS, 199+30)

    # ------------------
    #  positions des labels
    # TODO ramener depuis extérieur classe

    def __init__(self, ref_mod):
        super().__init__()
        fgcolor, bgcolor = pygame.color.Color(BIOS_FG_COL_DESC), pygame.color.Color(BIOS_BG_COL_DESC)

        self.bgcolor = bgcolor
        self.fgcolor = fgcolor

        self._ref_mod = ref_mod

        self._font = pygame.font.Font(None, 22)
        self._buttons = list()
        self._crea_boutons()

        # self.red_color = glvars.colors['c_cherokee']
        self._fin_aff_flottant = None

        gui.Etiquette.set_font(self._font)

        init_yv = 22
        y_lig1 = init_yv+4*self.VOFFSET
        y_lig2 = init_yv+5*self.VOFFSET
        self._etiquettes = [
            gui.Etiquette('KATA.GAMES auth procedure', (self.TAQ_XALIGN_LABELS, init_yv), self.fgcolor),
            gui.Etiquette('if you need a new account, please', (self.TAQ_XALIGN_LABELS, init_yv+1*self.VOFFSET), self.fgcolor),
            gui.Etiquette('use our website https://kata.games', (self.TAQ_XALIGN_LABELS, init_yv+2*self.VOFFSET), self.fgcolor),
            gui.Etiquette('username', (self.TAQ_XALIGN_LABELS, y_lig1), self.fgcolor),
            gui.Etiquette('password', (self.TAQ_XALIGN_LABELS, y_lig2), self.fgcolor)
        ]

        self._err_etiquette = gui.Etiquette(
            'ERR: server does not recognize this account', (self.TAQ_XALIGN_LABELS, init_yv+3*self.VOFFSET), (255, 0, 0)
        )
        self.saisie_txt1 = gui.TextInput('>', self._font, KataFrameV.nop, (self.TAQ_XALIGN_BOX, y_lig1), self.TEXT_INPUT_SIZE)
        self.saisie_txt2 = gui.TextInput('>', self._font, KataFrameV.nop, (self.TAQ_XALIGN_BOX, y_lig2), self.TEXT_INPUT_SIZE)
        self.saisie_txt2.pwd_field = True

    @staticmethod
    def confirm_routine():
        # TODO déclencher interrogation serveur
        # TODO exporter trucs vers le contrôleur, c'est lui qui va popstate
        ev = CgmEvent(EngineEvTypes.DOAUTH)
        EventManager.instance().post(ev)

    @staticmethod
    def cancel_routine():
        ev = CgmEvent(EngineEvTypes.POPSTATE)  # state_ident=GameStates.IntroJeu)
        EventManager.instance().post(ev)

    @staticmethod
    def nop(txt):
        pass

    def _crea_boutons(self):
        txt = 'connection'
        bt = gui.Button(self._font, txt, self.POS_BT_CONFIRM, KataFrameV.confirm_routine)
        self._buttons.append(bt)

        txt = 'abort'
        bt = gui.Button(self._font, txt, self.POS_BT_CANCEL, KataFrameV.cancel_routine)
        self._buttons.append(bt)

    # - activation / désactivation impacte boutons également
    def turn_on(self, prio=None):
        super().turn_on()

        for bt in self._buttons:
            bt.turn_on()

        if self._ref_mod.is_focusing_username():
            self.saisie_txt1.focus()
            self.saisie_txt1.turn_on()
        else:
            self.saisie_txt2.focus()
            self.saisie_txt2.turn_on()

    def turn_off(self):
        super().turn_off()

        for bt in self._buttons:
            bt.turn_off()

        if self._ref_mod.is_focusing_username():
            self.saisie_txt1.no_focus()
            self.saisie_txt1.turn_off()
        else:
            self.saisie_txt2.no_focus()
            self.saisie_txt2.turn_off()

    # ########################################################
    #  proc_event
    # ########################################################
    def proc_event(self, ev, source):
        if ev.type == EngineEvTypes.PAINT:
            self._paint(ev.screen)

        # GUI controls
        elif ev.type == pygame.MOUSEBUTTONDOWN:
            if self.saisie_txt1.contains(ev.pos):
                self._ref_mod.set_focus(KataFrameM.VAR_USERNAME)
            elif self.saisie_txt2.contains(ev.pos):
                self._ref_mod.set_focus(KataFrameM.VAR_PWD)

        elif ev.type == pygame.KEYDOWN:
            if ev.key == pygame.K_TAB:
                self._ref_mod.toggle_focus()
            elif ev.key == pygame.K_RETURN or ev.key == pygame.K_KP_ENTER:
                KataFrameV.confirm_routine()  # view > ctrl comm.

        # updating view (mod > view comm.)
        elif ev.type == EngineEvTypes.FIELDCH:
            if ev.field:  # pwd
                if ev.txt == '':
                    self.saisie_txt2.move_caret('home')
                    while self.saisie_txt2.get_disp_text() != '':
                        self.saisie_txt2.delete_char()
                    self.saisie_txt2.render_field()
            else:
                if ev.txt == '':
                    self.saisie_txt1.move_caret('home')
                    while self.saisie_txt1.get_disp_text() != '':
                        self.saisie_txt1.delete_char()
                    self.saisie_txt1.render_field()

        elif ev.type == EngineEvTypes.FOCUSCH:
            if ev.focusing_username:
                self._gui_focus_username()
            else:
                self._gui_focus_pwd()

    # - métier
    def _gui_focus_username(self):
        if self.saisie_txt2.has_focus():
            self.saisie_txt2.no_focus()
            self.saisie_txt2.turn_off()

            self.saisie_txt1.focus()
            self.saisie_txt1.turn_on()

    def _gui_focus_pwd(self):
        if self.saisie_txt1.has_focus():
            self.saisie_txt1.no_focus()
            self.saisie_txt1.turn_off()

            self.saisie_txt2.focus()
            self.saisie_txt2.turn_on()

    def _paint(self, screen):
        screen.fill(self.bgcolor)
        for e in self._etiquettes:
            screen.blit(e.img, e.pos)

        # - dessin boutons & widgets
        for b in self._buttons:
            screen.blit(b.image, b.position)
        screen.blit(self.saisie_txt1.image, self.saisie_txt1.position)
        screen.blit(self.saisie_txt2.image, self.saisie_txt2.position)

        if self._fin_aff_flottant is not None:
            t = time.time()
            if t > self._fin_aff_flottant:
                self._fin_aff_flottant = None
            else:
                screen.blit(self._err_etiquette.img, self._err_etiquette.pos)

    def sig_bad_auth(self):  # triggered externally by a ctrl class
        self._fin_aff_flottant = time.time() + self.ERR_MSG_DELAY



# Cribbed from the Red Blob Games tutorial.

import heapq


class PriorityQueue:
    def __init__(self):
        self.elements = []

    def empty(self):
        return len(self.elements) == 0

    def put(self, item, priority):
        heapq.heappush(self.elements, (priority, item))

    def get(self):
        return heapq.heappop(self.elements)[1]


class AStarPath(object):
    DELTA8 = ((-1, -1), (0, -1), (1, -1), (-1, 0), (1, 0), (-1, 1), (0, 1), (1, 1))
    #DELTA4 = ((0, -1), (-1, 0), (1, 0), (0, 1))

    def __init__(self, mymap, start, goal, blocked_fun, clamp_fun, wrap_x=False, wrap_y=False, blocked_tiles=()):
        # blocked_fun is a function that takes mymap, x, y and returns True if movement into that tile is blocked.y
        # clamp_fun is a function that takes (x,y) and clamps the values if needed.
        start = clamp_fun(start)
        self.start = start
        goal = clamp_fun(goal)
        self.goal = goal
        self.blocked_tiles = set(blocked_tiles)
        self.mymap = mymap
        self.blocked_fun = blocked_fun
        self.clamp_fun = clamp_fun
        self.wrap_x = wrap_x
        self.wrap_y = wrap_y
        frontier = PriorityQueue()
        frontier.put(self.start, 0)
        self.came_from = {}
        self.cost_to_tile = {}
        self.came_from[self.start] = None
        self.cost_to_tile[self.start] = 0

        while not frontier.empty():
            current = frontier.get()

            if current == goal:
                break

            for next in self.neighbors(mymap, current):
                new_cost = self.cost_to_tile[current] + self.movecost(current, next)
                if next not in self.cost_to_tile or new_cost < self.cost_to_tile[next]:
                    self.cost_to_tile[next] = new_cost
                    priority = new_cost + self.heuristic(goal, next)
                    frontier.put(next, priority)
                    self.came_from[next] = current

        self.results = self.get_path(goal)

    def get_path(self, goal):
        results = list()
        p = goal
        while p:
            results.append(p)
            p = self.came_from.get(p)
        results.reverse()
        if results[0] != self.start:
            results = list()
        return results

    def neighbors(self, mymap, pos):
        x, y = pos
        for dx, dy in self.DELTA8:
            #x2, y2 = x + dx, y + dy
            x2, y2 = self.clamp_fun((x + dx, y + dy))
            #x2, y2 = int(x2), int(y2)
            if not ((x2, y2) in self.blocked_tiles or self.blocked_fun(mymap, x2, y2)):
                yield (x2, y2)
            elif (x2, y2) == self.goal:
                yield self.goal

    def movecost(self, a, b):
        xcost = abs(a[0] - b[0])
        ycost = abs(a[1] - b[1])
        if self.wrap_x:
            xcost = min(xcost, self.mymap.width - xcost)
        if self.wrap_y:
            ycost = min(ycost, self.mymap.height - ycost)
        return 1 + xcost + ycost

    def heuristic(self, a, b):
        # Manhattan distance on a square grid
        dx = abs(a[0] - b[0])
        dy = abs(a[1] - b[1])
        if self.wrap_x:
            dx = min(dx, self.mymap.width - dx)
        if self.wrap_y:
            dy = min(dy, self.mymap.height - dy)
        return (dx + dy)





class Transition:
    '''
    classe modélisant la transition dun FSA,
    on suppose que le déclencheur est un mot-clé
    si le déclencheur est repéré dans les info. reçues,
    la transition est déclenchée
    '''
    
    def __init__(self, info_decl, label_etat_suiv ):
        self.declencheur = info_decl
        self.label_etat_suiv = label_etat_suiv

    def est_declenchee(self, info):
        if self.declencheur in info:
            return True
        return False

    def get_label_etat_suiv(self):
        return self.label_etat_suiv


class Etat:
    '''
    classe modélisant un état de FSA,
    un état peut être initial ou terminal
    un état exploite des info. reçues pour déclencher ou non ses transitions
    vers un autre état
    '''

    def __init__(self, label, li_transitions, est_init=False, est_term=False):
        self.label = label
        self.li_transitions = li_transitions
        self.est_init = est_init
        self.est_term = est_term

    def exploite_info(self, info):
        '''retourne le label de l'état suivant OU None,
        selon qu'une transition est déclenchée OU non'''
        if self.est_terminal():
            return None
        for transition in self.li_transitions:
            #print( transition.label_etat_suiv )
            if transition.est_declenchee( info):
                return transition.get_label_etat_suiv()
        return None
        
    def est_terminal(self):
        return self.est_term

    def est_initial(self):
        return self.est_init

    def get_label(self):
        return self.label


class Automate:
    '''
    classe modélisant un FSA pour Finite States Automaton
    concept très utilisé en IA dans les jeux commerciaux,
    malgré sa simplicité c'est la base du comportement des PNJs
    '''

    def __init__(self, li_etats):
        self.etats = dict()
        for etat in li_etats:
            self.etats[etat.get_label() ] = etat

        self.reset()

    def entree_info(self, info):
        '''permet le changement detat '''
        res = self.get_etat_courant().exploite_info(info)
        if res!=None:
            self.label_etat_courant = res

    def get_etat_courant(self):
        return self.etats[self.label_etat_courant]

    def reset(self):
        init_st_trouve = False
        #initialisation de l'etat courant
        for etat in self.etats.values():
            if etat.est_initial():
                if init_st_trouve: #doublon
                    raise Exception("definition FSA non valide, il ne peut y avoir qu'un état initial")
                self.label_etat_courant = etat.get_label()
                init_st_trouve = True




#--- essais definition

s1 =  Etat( 's1', [ Transition( 'hello', 's2') ] , True, False )
s2 = Etat('s2', [ Transition('news','s4') , Transition('bye','s3') ] )
s4 = Etat('s4', [Transition('ok','s2')] )
s3 = Etat('s3', [] , False, True )

fsa_exemple1 = Automate( [s1,s2,s3,s4]  )



# pour tester, on peut interagir avec le FSA directement dans la console python
# (dessiner le FSA sur papier pour bien le comprendre )
# voici la trace de mes propres tests :


# >>> fsa_exemple1.getEtatCourant().estInitial()
# True
# >>> fsa_exemple1.getEtatCourant().getLabel()
#'s1'

# >>> fsa_exemple1.entreeInfo('hello')
# >>> fsa_exemple1.entreeInfo('news')
# >>> fsa_exemple1.getEtatCourant().getLabel()
#'s4'
# >>> fsa_exemple1.entreeInfo('ok')
# >>> fsa_exemple1.getEtatCourant().getLabel()
# 's2'
# >>> fsa_exemple1.entreeInfo('bye')
# >>> fsa_exemple1.getEtatCourant().getLabel()
# 's3'

# >>> fsa_exemple1.entreeInfo('hello')
# >>> fsa_exemple1.entreeInfo('bye')
# >>> fsa_exemple1.entreeInfo('nimporte quoi')
# >>> fsa_exemple1.getEtatCourant().getLabel()
# 's3'
# >>> fsa_exemple1.getEtatCourant().estTerminal()
# True

# >>> fsa_exemple1.reset()
# >>> fsa_exemple1.getEtatCourant().estInitial()
# True

       

              
        

    

    



from ..foundation import defs


_vsurface = None
_vsurface_required = True

cached_pygame_mod = None  # init from outside when one calls kengi.bootstrap_e
special_flip = 0  # flag, set it to 1 when using web ctx
screen = None
stored_upscaling = 1
defacto_upscaling = None

# hopefully i will be able to simplify this:
ctx_emuvram = None
canvas_emuvram = None
canvas_rendering = None
real_pygamescreen = None
screen_rank = 1  # so we can detect whenever its required to update the var in the PAINT engine event


def set_upscaling(new_upscal_val):
    global stored_upscaling, _vsurface_required
    if stored_upscaling is not None:
        if int(stored_upscaling) != new_upscal_val:
            stored_upscaling = int(new_upscal_val)
            _vsurface_required = True


def flip():
    global _vsurface_required, _vsurface
    if _vsurface_required:
        # TODO
        pass

    if not special_flip:  # flag can be off if the extra blit/transform has to disabled (web ctx)
        realscreen = cached_pygame_mod.display.get_surface()
        if 1 == stored_upscaling:
            realscreen.blit(screen, (0, 0))
        else:
            cached_pygame_mod.transform.scale(screen, defs.STD_SCR_SIZE, realscreen)

    cached_pygame_mod.display.update()


# ------------------------------------
#   old code
# ------------------------------------
_curr_state = None
_loaded_states = dict()
init2_done = False
state_stack = None


def conv_to_vscreen(x, y):
    return int(x / defacto_upscaling), int(y / defacto_upscaling)


# def set_canvas_rendering(jsobj):
#     shared.canvas_rendering = jsobj
#
#
# def set_canvas_emu_vram(jsobj):
#     shared.canvas_emuvram = jsobj
#     shared.ctx_emuvram = jsobj.getContext('2d')


def set_realpygame_screen(ref_surf):
    global real_pygamescreen
    if real_pygamescreen:
        print('warning: set_realpygame_scneen called a 2nd time. Ignoring request')
        return
    real_pygamescreen = ref_surf


def set_virtual_screen(ref_surface):
    global screen, screen_rank, defacto_upscaling
    screen = ref_surface
    w = screen.get_size()[0]
    defacto_upscaling = 960/w
    screen_rank += 1


def get_screen():
    return screen


def proj_to_vscreen(org_screen_pos):
    return conv_to_vscreen(*org_screen_pos)



"""
all elements in this sub-modules are imported directly

via Injector
(dependency injection manager)

use the:

  import katagames_sdk.engine as kataen
  kataen.*

syntax
"""

# N.b need to import so setup.py can detect all files

from . import console
from . import rogue
from . import tabletop
from . import terrain



"""
(~*~ legacy code ~*~)
AUTHOR INFO
 Joseph H. cf. https://github.com/jwvhewitt

LICENSE INFO
 for the whole polarbear folder: Apache 2.0

Below are chunks extracted from the Polar Bear Game Engine.
This package contains some low-level graphics stuff needed to
create an isometric RPG style rules in Python. The idea is to
isolate the graphics handling from the code as much as possible,
so that if PyGame is replaced the interface shouldn't change
too much. Also, so that creating a new rules should be as simple
as importing this package.

Word wrapper taken from the PyGame wiki plus the list-printer
from Anne Archibald's GearHead Prime demo.
"""

import glob
import random
import weakref
from ... import _hub
from .general import default_border, render_text, wait_event, TIMEREVENT, INFO_GREEN
from .image import draw_text
# has been deleted, its included within .demolib now
# from . import rpgmenu
from . import image
from . import widgets

# disabled as this
# creates a problem (circular import? Cannot import from partially initialized polarbear)
# from . import dialogue

pygame = _hub.pygame




from ... import _hub as inj


pygame = inj.pygame



def create_txtblock_surf(text, ft_obj, alignment_f=0.0, debug=0, color=(0, 0, 0), bg_color=None):
    """
    goal = to produce a pygame surface
    append a pair (lbl, textRect) to render_list,one per line

    if bg_color ->None then transparent bg!!!
    
    """
    render_list = list()
    mw = mh = -1
    toth = 0
    
    for line in text.split('\n'):

        # CustomFt: crea surface, puis render...
        #text_lbl = pygame.Surface((ft_obj.width(line), 16),pygame.SRCALPHA)
        #ft_obj.render(line, text_lbl, (0,0))

        #pyg font
        text_lbl = ft_obj.render(line, False, color, bg_color)
        
        w, h = text_lbl.get_size()
        if w > mw:
            mw = w
        if h > mh:
            mh = h
        toth += h
        render_list.append(text_lbl)

    # create surf, that represents txtblock
    surf = pygame.Surface((mw, toth), pygame.SRCALPHA)
    
    # stack labels, to gen the txt block
    target_y = 0
    for lbl in render_list:
        curr_w = lbl.get_width()
        surf.blit(lbl, (alignment_f * (mw - curr_w), target_y))
        target_y += mh
    if debug:
        pygame.draw.rect(surf, 'red', ((0, 0), (mw - 1, toth - 1)), 1)
    return surf


class TextBlock(pygame.sprite.Sprite):
    ALIGN_LEFT, ALIGN_CENTER = 0, 1

    #def __init__(self, ft_name, ft_size, text='', color=(0, 0, 0), bg_color=(255, 255, 255)):
    def __init__(self, ft_obj, text='', color=(0, 0, 0), bg_color=(255, 255, 255)):
        super().__init__()
        #self.ft_obj = pygame.font.Font(ft_name, ft_size)
        self._color = color
        self._bg_color = bg_color
        self.ft_obj = ft_obj
        
        self._align = self.ALIGN_CENTER
        self._debug = 0
        self.rect = None  # no position for now
        if '' == text:
            self._text = ''
            self.image = None
        else:
            self._text = text
            self._refresh_img()

    def _refresh_img(self):
        if self.rect:  # has a position
            prev_pos = self.rect.center
        else:
            prev_pos = None

        if self._align == self.ALIGN_CENTER:
            ts = create_txtblock_surf(self._text, self.ft_obj, 0.5, self.debug, self._color)
            # ts = create_txtblock_surf(self._text, self.ft_obj, self._color, self._bg_color, 0.5, self.debug)
        else:
            ts = create_txtblock_surf(self._text, self.ft_obj, 0.0, self.debug, self._color)
            # ts = create_txtblock_surf(self._text, self.ft_obj, self._color, self._bg_color, 0.0, self.debug)
        self.image = ts
        self.rect = self.image.get_rect()
        # restore previous position
        if prev_pos:
            self.rect.center = prev_pos

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, v):
        self._debug = 1 if v else 0
        self._refresh_img()

    @property
    def text_align(self):
        return self._align

    @text_align.setter
    def text_align(self, v):
        if v in (self.ALIGN_LEFT, self.ALIGN_CENTER):
            self._align = v
            self._refresh_img()
        else:
            raise ValueError

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, val):
        self._text = val
        self._refresh_img()

    def draw(self, scr):
        if self.image:
            pos = self.rect.topleft
            scr.blit(self.image, pos)



import threading

# import urllib.parse as pp
#  from http.client import HTTPConnection
from katagames_sdk.capsule.networking.httpserver import HttpServer


class EnvoiAsynchrone2(threading.Thread):
    NB_INST_MAX = 16

    nb_courant_instances = 0

    PORT = None
    HOST_STR = None
    
    def __init__(self, id_req, http_ressource, opt_data, test_timestamp=None):
        cls = self.__class__

        if cls.nb_courant_instances >= cls.NB_INST_MAX:
            raise Exception('Nb. maximum d\' appels asynchrones atteint')

        self.id_req = id_req  # utile pour fournir reponse via CgmEvent !!

        # self.test_timestamp = '1523728737' if (test_timestamp is None) else test_timestamp

        threading.Thread.__init__(self)
        cls.nb_courant_instances += 1

        self.post_data = opt_data  # peut être None si on fait du get, ou set correctement pour du POST

        # - TEMPORAIRE
        # self.server_port = self.__class__.PORT
        # self.host = self.__class__.HOST_STR

        # pr long-polling
        # self.url_pr_server_time = None
        
        self.target_http_ressource = http_ressource
        self.zombie_thread = threading.Event() 
        self.response = None

        # self.cb = None

    # def setup_callback(self, cb):
    #     self.cb = cb

    def __del__(self):
        self.__class__.nb_courant_instances -= 1

    # était utilisé pour le long polling...
    # def stop(self):
    #    pass
        # self.cli_http.close()

    # def f_long_polling_packet(self, ts_arg):
    #     if self.server_port is None:  # si non initialisé
    #         self.server_port = 80
    #         self.host = '192.168.1.99'
    #         self.url_pr_server_time = '/tom/server.php'
    #         self.debug_mode = True
    #
    #         #self.cli_http = HTTPConnection(self.host)
    #         #self.cli_http.connect()
    #
    #     # transmission
    #     msg = self.url_pr_server_time + '?timestamp={}'.format(ts_arg)
    #
    #     res = EnvoiAsynchrone2._commit_to_netw(self.host, msg)
    #     return res

    def run(self):
        url = 'http://{}:{}/{}'.format(EnvoiAsynchrone2.HOST_STR, EnvoiAsynchrone2.PORT, self.target_http_ressource)

        if self.post_data is not None:
            self.response = HttpServer.instance().proxied_post(url, {'data': self.post_data})
        else:
            self.response = HttpServer.instance().proxied_get(url, {})

        self.zombie_thread.set()  # indique que la tâche du thread est terminée



import random as _rand


class _Pal:
    def __init__(self, pal_tuple):
        self.ctable = dict()
        self.listing = list()
        self.names = list()
        for elt in pal_tuple:
            t3 = (elt[1], elt[2], elt[3])
            self.ctable[elt[0]] = t3
            self.listing.append(t3)
            self.names.append(elt[0])

        self._size = len(pal_tuple)

    @property
    def size(self):
        return self._size

    def get_name(self, rgb):
        for nam, desc in self.ctable.items():
            if rgb[0] == desc[0] and rgb[1] == desc[1] and rgb[2] == desc[2]:
                return nam

    def get_rank(self, rgb):
        return self.listing.index(rgb)

    def at_random(self, excl_set=None):
        """
        excl_set can be {0, 3, 8} for example,
        but can also be {'black', 'white', 'orange', 'blue'} for example
        """
        picknumber, exclusion = True, False
        if excl_set:
            if len(excl_set) > self._size:
                raise ValueError('excl_set cannot exclude that many colours!')

            e = None
            for e in excl_set:
                exclusion = True
                break
            if exclusion:
                if not isinstance(e, int):
                    picknumber = False
        else:
            excl_set = set()

        if picknumber:
            omega = list(range(self._size))
            for exck in excl_set:
                omega.remove(exck)
            return self.listing[_rand.choice(omega)]

        omega = list(self.names)
        for exck in excl_set:
            omega.remove(exck)
        return self.ctable[_rand.choice(omega)]

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.ctable[item]
        else:
            return self.listing[item]


# ------------------
#  Built-in palettes
# ------------------
_c64pdef = (
    ('black', 0, 0, 0),
    ('white', 255, 255, 255),
    ('red', 138, 20, 22),
    ('cyan', 120, 170, 207),

    ('violet', 204, 68, 204),
    ('green', 100, 184, 70),
    ('blue', 69, 32, 195),
    ('yellow', 207, 207, 101),

    ('orange', 154, 101, 32),
    ('brown', 101, 69, 0),
    ('lightred', 207, 138, 101),
    ('darkgrey', 69, 69, 69),

    ('grey', 138, 138, 138),
    ('lightgreen', 170, 239, 138),
    ('lightblue', 160, 162, 255),
    ('lightgrey', 170, 170, 170),
)

c64 = _Pal(_c64pdef)  # Commodore 64

_cyberpunkdef = (
    ('black', 0, 0, 0),
    ('nightblue', 1, 1, 42),
    ('darkpurple', 50, 20, 80),
    ('darkblue', 40, 40, 68),
    ('gray', 88, 82, 104),
    ('lightblue', 4, 218, 246),  # ('cyan', 115, 255, 254),
    ('cyangreen', 160, 253, 226),
    ('brightgreen', 0, 255, 65),
    ('lightgreen', 100, 221, 153),
    ('green', 106, 126, 105),
    ('brown', 85, 75, 65),
    ('lightbrown', 148, 89, 10),
    ('yellow', 243, 230, 0),
    ('glowinyellow', 222, 254, 71),
    ('paleyellow', 255, 230, 157),
    ('orange', 255, 110, 39),
    ('paleorange', 255, 150, 96),
    ('mud', 53, 21, 9),
    ('bloodstain', 83, 1, 9),
    ('red', 255, 0, 60),
    ('flashypink', 251, 16, 163),
    ('pink', 248, 135, 255),
    ('lightgray', 183, 192, 221),
    ('white', 227, 222, 223)
)

punk = _Pal(_cyberpunkdef)

_japan = (
    ('black', 0, 0, 0),
    ('bluegray', 43, 51, 95),
    ('burgundy', 126, 32, 114),
    ('cyan', 25, 149, 156),
    ('brown', 139, 72, 82),
    ('blue', 57, 92, 152),
    ('skyblue', 169, 192, 255),
    ('white', 238, 238, 238),
    ('red', 212, 24, 108),
    ('orange', 211, 132, 65),
    ('yellow', 233, 195, 91),
    ('green', 112, 198, 169),
    ('paleblue', 118, 150, 222),
    ('lightgray', 163, 163, 163),
    ('pink', 255, 151, 152),
    ('biscuit', 237, 199, 176)
)

japan = _Pal(_japan)

ALL_PALETTES = {
    'c64': c64,
    'japan': japan,
    'punk': punk,
}



import weakref
from itertools import chain

from ... import _hub
from ...compo import vscreen as core


DEFAULT_COLOR_KEY = (255, 0, 255)
TEXT_COLOR = (240, 240, 50)
pygame = _hub.pygame
# Keep a list of already-loaded images, to save memory when multiple objects
# need to use the same image file.
pre_loaded_images = weakref.WeakValueDictionary()


def truncline(text, font, maxwidth):
    real = len(text)
    stext = text
    l = font.size(text)[0]
    cut = 0
    a = 0
    done = 1
    old = None
    while l > maxwidth:
        a = a + 1
        n = text.rsplit(None, a)[0]
        if stext == n:
            cut += 1
            stext = n[:-cut]
        else:
            stext = n
        l = font.size(stext)[0]
        real = len(stext)
        done = 0
    return real, done, stext


def wrapline(text, font, maxwidth):
    done = 0
    wrapped = []
    while not done:
        nl, done, stext = truncline(text, font, maxwidth)
        wrapped.append(stext.strip())
        text = text[nl:]
    return wrapped


def wrap_multi_line(text, font, maxwidth):
    """ returns text taking new lines into account.
    """
    lines = chain(*(wrapline(line, font, maxwidth) for line in text.splitlines()))
    return list(lines)


def render_text(font, text, width, color=TEXT_COLOR, justify=-1, antialias=True):
    # Return an image with prettyprinted text.
    lines = wrap_multi_line(text, font, width)

    imgs = [font.render(l, antialias, color) for l in lines]
    h = sum(i.get_height() for i in imgs)
    s = pygame.surface.Surface((width, h))
    s.fill((0, 0, 0))
    o = 0
    for i in imgs:
        if justify == 0:
            x = width // 2 - i.get_width() // 2
        elif justify > 0:
            x = width - i.get_width()
        else:
            x = 0
        s.blit(i, (x, o))
        o += i.get_height()
    s.set_colorkey((0, 0, 0), pygame.RLEACCEL)
    return s


def draw_text(font, text, rect, color=TEXT_COLOR, justify=-1, antialias=True, dest_surface=None):
    # Draw some text to the screen with the provided options.
    if dest_surface:
        dsu = dest_surface
    else:
        dsu = core.get_screen()

    myimage = render_text(font, text, rect.width, color, justify, antialias)
    if justify == 0:
        myrect = myimage.get_rect(midtop=rect.midtop)
    elif justify > 0:
        myrect = myimage.get_rect(topleft=rect.topleft)
    else:
        myrect = rect
    dsu.set_clip(rect)
    dsu.blit(myimage, myrect)
    dsu.set_clip(None)


class Image(object):
    def __init__(self, fname=None, frame_width=0, frame_height=0, custom_frames=None,
                 flags=0, color_key=DEFAULT_COLOR_KEY, transparent=False):
        """Load image file, or create blank image, at frame size"""
        if fname:
            self.bitmap = self.get_pre_loaded(fname, transparent)
            if not self.bitmap:
                self.bitmap = pygame.image.load(fname).convert_alpha()
                self.bitmap.set_colorkey(color_key, flags)
        else:
            self.bitmap = pygame.Surface((frame_width, frame_height))
            self.bitmap.fill(color_key)
            self.bitmap.set_colorkey(color_key, flags)

        self.scrref = core.get_screen()

        self.fname = fname
        self.flags = flags
        self.transparent = transparent
        self.color_key = color_key
        if transparent:
            alpha = int(transparent)
            if alpha <= 1:
                alpha = 155
            self.bitmap.set_alpha(alpha)

        if frame_width == 0:
            frame_width = self.bitmap.get_width()
        if frame_height == 0:
            frame_height = self.bitmap.get_height()

        if frame_width > self.bitmap.get_width():
            frame_width = self.bitmap.get_width()
        self.fname = fname
        self.frame_width = frame_width
        self.frame_height = frame_height

        self.custom_frames = custom_frames

    @staticmethod
    def get_pre_loaded(ident, transparent):
        return pre_loaded_images.get((ident, transparent))

    @staticmethod
    def record_pre_loaded(ident, colorset, bitmap, transparent=False):
        pre_loaded_images[(ident, repr(colorset), transparent)] = bitmap

    def _get_frame_area(self, frame):
        if self.custom_frames and frame < len(self.custom_frames):
            area = pygame.Rect(self.custom_frames[frame])
        else:
            frames_per_row = self.bitmap.get_width() // self.frame_width
            area_x = (frame % frames_per_row) * self.frame_width
            area_y = (frame // frames_per_row) * self.frame_height
            area = pygame.Rect(area_x, area_y, self.frame_width, self.frame_height)
        return area

    def render(self, dest=(0, 0), frame=0, dest_surface=None ):
        # Render this Image onto the provided surface.
        # Start by determining the correct sub-area of the image.
        area = self._get_frame_area(frame)
        dest_surface = dest_surface or self.scrref  # new way to retrieve the surface used for display

        dest_surface.blit(self.bitmap, dest, area)

    def render_c(self, dest=(0, 0), frame=0, dest_surface=None ):
        # As above, but the dest coordinates point to the center of the image.
        area = self._get_frame_area(frame)

        dest_c = self.get_rect(frame)
        dest_c.center = dest
        dest_surface = dest_surface or core.get_screen()  # new way to retrieve the surface used for display

        dest_surface.blit(self.bitmap, dest_c, area)

    def get_subsurface(self, frame):
        # Return one of the frames as a pygame subsurface.
        area = self._get_frame_area(frame)
        return self.bitmap.subsurface(area)

    def get_rect(self, frame):
        # Return a rect of the correct size for this frame.
        if self.custom_frames and frame < len(self.custom_frames):
            return pygame.Rect(0, 0, self.custom_frames[frame][2], self.custom_frames[frame][3])
        else:
            return pygame.Rect(0, 0, self.frame_width, self.frame_height)

    def num_frames(self):
        if self.custom_frames:
            return len(self.custom_frames)
        else:
            frames_per_row = self.bitmap.get_width() // self.frame_width
            frames_per_column = self.bitmap.get_height() // self.frame_height
            return frames_per_row * frames_per_column

    def __reduce__(self):
        # Rather than trying to save the bitmap image, just save the parameters.
        return Image, (self.fname, self.frame_width, self.frame_height, self.custom_frames, self.flags,
                       self.color_key, self.transparent)

    def tile(self, dest=None, frame=0, dest_surface=None, x_offset=0, y_offset=0):
        dest_surface = dest_surface or self.scrref  # new way to retrieve the surface used for display

        if not dest:
            dest = dest_surface.get_rect()
        grid_w = dest.w // self.frame_width + 2
        grid_h = dest.h // self.frame_height + 2
        dest_surface.set_clip(dest)
        my_rect = pygame.Rect(0, 0, 0, 0)

        for x in range(-1, grid_w):
            my_rect.x = dest.x + x * self.frame_width +x_offset
            for y in range(-1, grid_h):
                my_rect.y = dest.y + y * self.frame_height +y_offset
                self.render(my_rect, frame, dest_surface)

        dest_surface.set_clip(None)

    def copy(self,ident=None):
        nu_sprite = Image(frame_height=self.frame_height,frame_width=self.frame_width,)
        nu_sprite.bitmap = self.bitmap.copy()
        if ident:
            self.record_pre_loaded(ident, None, nu_sprite.bitmap)
        return nu_sprite


class TextImage(Image):
    def __init__(self, txt='?????', frame_width=128, color=None, font=None):
        """Create an image of the provided text"""
        if not color:
            color = TEXT_COLOR

        self.txt = txt
        self.bitmap = render_text(font, txt, frame_width, color, justify=0, antialias=False)
        self.frame_width = self.bitmap.get_width()
        self.frame_height = self.bitmap.get_height()
        self.custom_frames = None

    def __reduce__(self):
        # Rather than trying to save the bitmap image, just save the reconstruction details.
        return TextImage, (self.txt, self.frame_width, self.frame_height)





import katagames_sdk.capsule.engine_ground.conf_eng as cgmconf
from katagames_sdk.capsule.event import EventReceiver, EngineEvTypes
from katagames_sdk.capsule.pygame_provider import get_module

pyg = get_module()


class MicroView(EventReceiver):
    def __init__(self):
        super().__init__()
        fgcolor, self.bgcolor = pyg.color.Color('purple'), pyg.color.Color('antiquewhite3')

        self._font = pyg.font.Font(None, 22)

        self._label = self._font.render('CLICK HERE TO PLAY!', False, fgcolor)
        scr_w, scr_h = cgmconf.get_screen().get_size()
        self._pos = (
            (scr_w - self._label.get_width())//2,
            (scr_h - self._label.get_height())//2
        )

    def proc_event(self, ev, source):
        if ev.type == EngineEvTypes.PAINT:
            ev.screen.fill(self.bgcolor)
            ev.screen.blit(self._label, self._pos)

        elif ev.type == pyg.MOUSEBUTTONUP:
            self.pev(EngineEvTypes.CHANGESTATE, state_ident=0)



from ... import _hub as inj
from ... import event


pygame = inj.pygame


def test_func():
    print("Button clicked!")


class Button2(event.EventReceiver):
    HIDEOUS_PURPLE = (255, 0, 255)

    def __init__(self, font, text, position_on_screen, callback=None, draw_background=True):
        super().__init__()
        padding_value = 12  # pixels

        if draw_background:
            self.bg_color = (100, 100, 100)
            self.fg_color = (255, 255, 255)
        else:
            self.bg_color = self.HIDEOUS_PURPLE
            self.fg_color = (0, 0, 0)

        # data
        self._callback = callback
        self._text = text
        self._hit = False

        # dawing
        self.font = font
        self.txt = text

        size = font.size(text)
        self.tmp_size = size
        self.position = position_on_screen
        self.collision_rect = pygame.Rect(self.position, size).inflate(padding_value, padding_value)
        self.collision_rect.topleft = self.position

        self.image = None
        self.refresh_img()

    def refresh_img(self):
        self.image = pygame.Surface(self.collision_rect.size).convert()
        self.image.fill(self.bg_color)

        if self.bg_color != self.HIDEOUS_PURPLE:
            textimg = self.font.render(self.txt, True, self.fg_color, self.bg_color)
        else:
            textimg = self.font.render(self.txt, False, self.fg_color)

        ssdest = self.image.get_size()
        ssource = textimg.get_size()
        blit_dest = (
            (ssdest[0] - ssource[0]) // 2,
            (ssdest[1] - ssource[1]) // 2
        )
        self.image.blit(textimg, blit_dest)

        #  TODO is this useful?
        # if self.bg_color == self.HIDEOUS_PURPLE:
        #     self.image.set_colorkey((self.bg_color))
        #     box_color = (190,) * 3
        #     full_rect = (0, 0, self.image.get_size()[0], self.image.get_size()[1])
        #     pygame.draw.rect(self.image, box_color, full_rect, 1)

    # pour des raisons pratiques (raccourci)
    def get_size(self):
        return self.image.get_size()

    def proc_event(self, ev, source):
        if ev.type == pygame.KEYDOWN:
            self.on_keydown(ev)
        elif ev.type == pygame.MOUSEMOTION:
            self.on_mousemotion(ev)
        elif ev.type == pygame.MOUSEBUTTONDOWN:
            self.on_mousedown(ev)
        elif ev.type == pygame.MOUSEBUTTONUP:
            self.on_mouseup(ev)

    def on_keydown(self, event):
        """
        Decides what do to with a keypress.
        special meanings have these keys:
        enter, left, right, home, end, backspace, delete
        """
        if event.type != pygame.KEYDOWN:
            print("textentry got wrong event: " + str(event))
        else:
            self.render()

    ### debug
    # if __name__=='__main__' and event.key == pygame.K_ESCAPE:
    #     events.RootEventSource.instance().stop()

    def on_mousedown(self, event):
        pos = event.pos
        if self.collision_rect.collidepoint(pos):
            if self._callback:
                self._callback()

    def on_mouseup(self, event):
        pos = event.pos
        # print('button: mouse button up detected! x,y = {}'.format(pos))

    def on_mousemotion(self, event):
        pass

    def set_callback(self, callback):
        self._callback = callback

    def render(self):
        """

        """
        pass

    def update(self):
        pass
    #     """
    #     Actually not needed. (only need if this module is run as a script)
    #     """
    #     # only need if this module is run as a script
    #     if __name__ == '__main__':
    #         screen = pygame.display.get_surface()
    #         screen.fill((100, 0, 0))
    #         screen.blit(self.image, self.position)
    #         pygame.display.flip()




import base64
import io
import json
from collections import defaultdict

from . import packed_capello_ft
from .. import _hub


class JsonBasedSprSheet:
    def __init__(self, required_infos, ck=None):
        if isinstance(required_infos, str):
            filename_no_ext = required_infos
            self.sheet_surf = _hub.pygame.image.load(filename_no_ext+'.png')
            json_def_file = open(required_infos + '.json', 'r')
        else:
            fptr, json_def_file = required_infos
            self.sheet_surf = _hub.pygame.image.load(fptr)  # namehint="capello-ft.png"

        if ck:
            self.sheet_surf.set_colorkey(ck)

        jsondata = json.load(json_def_file)
        assoc_tmp = dict()
        self.all_names = set()
        if isinstance(jsondata['frames'], list):  # we support 2 formats of json desc
            for infos in jsondata['frames']:
                gname = infos['filename']
                self.all_names.add(gname)
                args = (infos['frame']['x'], infos['frame']['y'], infos['frame']['w'], infos['frame']['h'])
                assoc_tmp[gname] = self.sheet_surf.subsurface(_hub.pygame.Rect(*args)).copy()
        else:
            for sprname, infos in jsondata['frames'].items():
                self.all_names.add(sprname)
                args = (infos['frame']['x'], infos['frame']['y'], infos['frame']['w'], infos['frame']['h'])
                assoc_tmp[sprname] = self.sheet_surf.subsurface(_hub.pygame.Rect(*args)).copy()
        self.assoc_name_spr = assoc_tmp

    def __getitem__(self, item):
        return self.assoc_name_spr[item]


class ProtoFont:
    UNKNOWN_CAR_RK = 123
    SPAM_CAR = False

    def __init__(self, font_source=None):

        if font_source:  # meth1 : on ouvre des fichiers normalement
            print('[ProtoFont] using the provided pair of files:', font_source)
            self._sheet = JsonBasedSprSheet(
                font_source, ck=(127, 127, 127)  # font_source could be 'capello-ft' for example
            )
        else:  # default embedded caracter set
            # - meth2 : on ouvre du packed data
            filelike_png = io.BytesIO(base64.b64decode(packed_capello_ft.pngdata))
            filelike_json = io.StringIO(packed_capello_ft.jsondata)
            self._sheet = JsonBasedSprSheet(
                (filelike_png, filelike_json), ck=(127, 127, 127)
            )

        # specific to capello-ft.png and capello-ft.json...
        # it maps ascii codes to the rank font000.png where 000 is the rank
        mappingtable = {
            174: 150,  # circled r car.
            175: 151,
        }
        for e in range(32, 48):
            mappingtable[e] = e - 32
        for e in range(48, 64):
            mappingtable[e] = e - 31
        for e in range(64, 80):
            mappingtable[e] = e - 30
        for e in range(80, 96):
            mappingtable[e] = e - 29
        for e in range(96, 112):
            mappingtable[e] = e - 28
        for e in range(112, 127):
            mappingtable[e] = e - 27
        # upside-down !, cent ¢ , then £ symbol ... etc.
        for e in range(160, 173):
            mappingtable[e] = e - 24

        for e in range(176, 176+16):
            mappingtable[e] = e - 23
        for e in range(192, 208):
            mappingtable[e] = e - 22
        # Ð 208 et suivants
        for e in range(208, 224):
            mappingtable[e] = e - 21
        for e in range(224, 240):
            mappingtable[e] = e - 20
        # ð 240 et suivants
        for e in range(240, 256):
            mappingtable[e] = e - 19
        self.car_height = defaultdict(lambda: 7)
        # for my_asciicode in range(*alphabet_span):
        #     self.car_height[chr(my_asciicode)] = 7  # CONST

        # - generic
        self.ascii2img = dict()
        defaultw = self._sheet['tile{:03d}.png'.format(self.UNKNOWN_CAR_RK)].get_width()
        self.car_width = defaultdict(lambda: defaultw)

        for my_asciicode in mappingtable.keys():
            ssurf = self._sheet['tile{:03d}.png'.format(mappingtable[my_asciicode])]
            self.ascii2img[my_asciicode] = ssurf
            self.car_width[chr(my_asciicode)] = ssurf.get_width()

    @property
    def sheet(self):
        return self._sheet

    def __getitem__(self, itemk):
        ascii_tmp = ord(itemk)
        if self.SPAM_CAR:
            print(itemk, ascii_tmp)
        try:
            return self.ascii2img[ascii_tmp]
        except KeyError:
            return self._sheet['tile{:03d}.png'.format(self.UNKNOWN_CAR_RK)]

    def text_to_surf(self, w, refsurf, start_pos, spacing=0, bgcolor=None):
        # fill background with a solid color, if requested
        if bgcolor:
            curr_pos = list(start_pos)
            h = float('-inf')
            for letter in w:
                curr_pos[0] += self.car_width[letter] + spacing
                if self.car_height[letter] > h:
                    h = self.car_height[letter]
            _hub.pygame.draw.rect(
                refsurf, bgcolor, (start_pos[0], start_pos[1], curr_pos[0]-spacing-start_pos[0], h), 0
            )
        # draw the text
        curr_pos = list(start_pos)
        for letter in w:
            refsurf.blit(self[letter], curr_pos)
            curr_pos[0] += self.car_width[letter] + spacing

    def compute_width(self, w, spacing=0):
        res = 0
        for letter in w:
            res += self.car_width[letter] + spacing
        return res


class Spritesheet:
    """
    handles sprite sheets in an optimized way!
    REMARK: When calling images_at the rect is the format: (x, y, x + offset, y + offset)
    """

    def __init__(self, resource_info, chosen_scale=1):
        """
        :param resource_info: either pygame.Surface or filepath
        :param chosen_scale:
        """
        if isinstance(resource_info, _hub.pygame.Surface):
            self._sheet = resource_info
        else:
            self._sheet = _hub.pygame.image.load(resource_info).convert()

        if float(chosen_scale) != 1.0:
            homo = _hub.pygame.transform.scale
            w, h = self._sheet.get_size()
            self._sheet = homo(self._sheet, (chosen_scale*w, chosen_scale*h))

        print('** SHEET size ** ', self._sheet.get_size())

        # can be init later
        self._per_line_img_quant = 0
        self._size = 0
        self._tilesize = None

        self.colorkey = None

        # goal: speed-up
        self.cache = defaultdict(lambda: None)

    @property
    def card(self):
        return self._size

    @property
    def tilesize(self):
        return self._tilesize

    @tilesize.setter
    def tilesize(self, pair_wh):
        print('** tilesize set **', pair_wh)
        self._tilesize = tw, th = pair_wh
        # compute the nb of img and how many img per line
        self._per_line_img_quant = self._sheet.get_width() // tw
        self._size = self._per_line_img_quant * (self._sheet.get_height() // th)
        self.cache.clear()

    def __getitem__(self, item):
        return self.image_by_rank(item)

    def image_at(self, rectangle):
        y = self._sheet.subsurface(rectangle).copy()
        if self.colorkey:
            y.set_colorkey(self.colorkey)
        return y

    # USE WITH CAUTION! This method provides no optimization
    def images_at(self, rects, colorkey):
        """
        Loads a bunch of images at once
        :param rects: a list of coordinates
        :param colorkey:
        :return: several images as a list
        """
        res = [self.image_at(rect) for rect in rects]
        for e in res:
            e.set_colorkey(colorkey)
        return res

    def image_by_rank(self, kval):
        if self._tilesize is None:
            raise ValueError('Spritesheet.image_by_rank call but tilesize hasnt been set!')
        y = self.cache[kval]
        if y is None:
            tw, th = self._tilesize
            # map kval -> to a rect
            i, j = kval % self._per_line_img_quant, int(kval / self._per_line_img_quant)
            rect_obj = _hub.pygame.Rect(i*self._tilesize[0], j*self._tilesize[1], tw, th)
            # crop from the sheet save & return the result
            y = self.cache[kval] = self.image_at(rect_obj)
        return y

    # USE WITH CAUTION! This method provides no optimization
    def load_strip(self, rect_img0, image_count, colorkey=None):
        """Loads a strip of images and returns them as a list, rect must cut out the img rank 0"""
        rect = rect_img0
        tw, th = rect[2], rect[3]
        tups = [(rect[0] + x*tw, rect[1], tw, th) for x in range(image_count)]
        return self.images_at(tups, colorkey)




SCROLL_STEP = 0.25
FLIPPED_HORIZONTALLY_FLAG = 0x80000000
FLIPPED_VERTICALLY_FLAG = 0x40000000
FLIPPED_DIAGONALLY_FLAG = 0x20000000
ROTATED_HEXAGONAL_120_FLAG = 0x10000000
NOT_ALL_FLAGS = 0x0FFFFFFF



from katagames_sdk.capsule.event import CogObject
from katagames_sdk.capsule.event import EngineEvTypes


class KataFrameM(CogObject):
    """
    classe utilisée pr stocker temporairement login / pwd
    durant le processus de login
    Le pwd est gardé secret (pas d'affichage) mais on compte et
    on affiche correctement le nb de car
    """
    VAR_USERNAME, VAR_PWD = range(87, 87 + 2)

    def __init__(self):
        super().__init__()
        self._username_info = ''
        self._pwdplain_info = ''

        self._focusing_username = True

    def input_char(self, c):
        if self._focusing_username:
            self._username_info += c
        else:
            self._pwdplain_info += c

    # - getters
    @property
    def username(self):
        return self._username_info

    @property
    def plainpwd(self):
        return self._pwdplain_info

    def get_pwd_str(self):
        k = len(self._pwdplain_info)
        li_etoiles = '*' * k
        return ''.join(li_etoiles)

    def is_focusing_username(self):
        return self._focusing_username

    # - misc methods
    def set_focus(self, var_code):

        if var_code == self.VAR_USERNAME:
            if not self._focusing_username:
                self._focusing_username = True
                self.pev(EngineEvTypes.FOCUSCH, focusing_username=True)

        elif var_code == self.VAR_PWD:
            if self._focusing_username:
                self._focusing_username = False
                self.pev(EngineEvTypes.FOCUSCH, focusing_username=False)
        else:
            raise ValueError('unrecognized var_code')

    def toggle_focus(self):
        self._focusing_username = not self._focusing_username
        self.pev(EngineEvTypes.FOCUSCH, focusing_username=self._focusing_username)

    def suppr_lettre(self):
        if self._focusing_username:
            if len(self._username_info) > 0:
                self._username_info = self._username_info[:-1]
                self.pev(EngineEvTypes.FIELDCH, field=0, txt=self._username_info)
        else:
            if len(self._pwdplain_info) > 0:
                self._pwdplain_info = self._pwdplain_info[:-1]
                self.pev(EngineEvTypes.FIELDCH, field=1, txt=self.get_pwd_str())

    def reset_fields(self):
        self._username_info = ''
        self._pwdplain_info = ''

        self.pev(EngineEvTypes.FIELDCH, field=0, txt=self._username_info)
        self.pev(EngineEvTypes.FIELDCH, field=1, txt=self.get_pwd_str())

        self._focusing_username = True
        self.pev(EngineEvTypes.FOCUSCH, focusing_username=True)



from ... import _hub
import json
import os
from xml.etree import ElementTree
from .flags import *
from zlib import decompress
from base64 import b64decode
import struct
import math


pygame = _hub.pygame
Tilesets = _hub.tmx.data.Tilesets

info_type_obj = 'class'


class IsometricTile:
    def __init__(self, tile_id, tile_surface, hflip, vflip):
        self.id = tile_id
        self.tile_surface = tile_surface

        self.hflip_surface = pygame.transform.flip(tile_surface, True, False).convert_alpha()
        self.hflip_surface.set_colorkey(tile_surface.get_colorkey(), tile_surface.get_flags())
        self.vflip_surface = pygame.transform.flip(tile_surface, False, True).convert_alpha()
        self.vflip_surface.set_colorkey(tile_surface.get_colorkey(), tile_surface.get_flags())
        self.hvflip_surface = pygame.transform.flip(tile_surface, True, True).convert_alpha()
        self.hvflip_surface.set_colorkey(tile_surface.get_colorkey(), tile_surface.get_flags())

    def paint_tile(self, dest_surface, x, y, hflip=False, vflip=False):
        """Draw this tile on the dest_surface at the provided x,y coordinates."""
        if hflip and vflip:
            surf = self.hvflip_surface
        elif hflip:
            surf = self.hflip_surface
        elif vflip:
            surf = self.vflip_surface
        else:
            surf = self.tile_surface
        mydest = surf.get_rect(midbottom=(x, y))
        dest_surface.blit(surf, mydest)

    def __repr__(self):
        return '<Tile {}>'.format(self.id)


class IsometricTileset:
    """
    Based on the Tileset class from katagames_engine/_sm_shelf/tmx/data.py, but modified for the needs of isometric
    maps. Or at least the needs of this particular isometric map.
    """

    def __init__(self, name, tile_width, tile_height, firstgid):
        self.name = name
        self.tile_width = tile_width
        self.tile_height = tile_height
        self.firstgid = firstgid

        self.hflip = False
        self.vflip = False

        self.tiles = []
        self.properties = {}

    def get_tile(self, gid):
        return self.tiles[gid - self.firstgid]

    def _add_image(self, folders, source, num_tiles):
        # TODO: Make this bit compatible with Kenji.
        mysurf = pygame.image.load(os.path.join(os.sep.join(folders), source)).convert_alpha()
        mysurf.set_colorkey((255, 0, 255))
        myrect = pygame.Rect(0, 0, self.tile_width, self.tile_height)
        frames_per_row = mysurf.get_width() // self.tile_width

        for frame in range(num_tiles):
            myrect.x = (frame % frames_per_row) * self.tile_width
            myrect.y = (frame // frames_per_row) * self.tile_height
            self.tiles.append(IsometricTile(frame + 1, mysurf.subsurface(myrect), self.hflip, self.vflip))

    @classmethod
    def fromxml(cls, folders, tag, firstgid=None):
        print('fromxml (isometrically)')
        if 'source' in tag.attrib:
            # Instead of a tileset proper, we have been handed an external tileset tag from inside a map file.
            # Load the external tileset and continue on as if nothing had happened.
            firstgid = int(tag.attrib['firstgid'])
            srcc = tag.attrib['source']

            # TODO: Another direct disk access here.
            if srcc.endswith(("tsx", "xml")):
                with open(os.path.join(os.pathsep.join(folders), srcc)) as f:
                    print('opened ', srcc)
                    tag = ElementTree.fromstring(f.read())
            elif srcc.endswith(("tsj", "json")):
                with open(os.path.join(os.pathsep.join(folders), srcc)) as f:
                    jdict = json.load(f)
                return cls.fromjson(folders, jdict, firstgid)

        name = tag.attrib['name']
        if firstgid is None:
            firstgid = int(tag.attrib['firstgid'])
        tile_width = int(tag.attrib['tilewidth'])
        tile_height = int(tag.attrib['tileheight'])
        num_tiles = int(tag.attrib['tilecount'])

        tileset = cls(name, tile_width, tile_height, firstgid)

        # TODO: The transformations must be registered before any of the tiles. Is there a better way to do this
        # than iterating through the list twice? I know this is a minor thing but it bothers me.
        for c in tag:  # .getchildren():
            if c.tag == "transformations":
                tileset.vflip = int(c.attrib.get("vflip", 0)) == 1
                tileset.hflip = int(c.attrib.get("hflip", 0)) == 1
                print("Flip values: v={} h={}".format(tileset.vflip, tileset.hflip))

        for c in tag:  # .getchildren():
            # TODO: The tileset can only contain an "image" tag or multiple "tile" tags; it can't combine the two.
            # This should be enforced. For now, I'm just gonna support spritesheet tiles.
            if c.tag == "image":
                # create a tileset
                arg_sheet = c.attrib['source']
                tileset._add_image(folders, arg_sheet, num_tiles)

        return tileset

    @classmethod
    def fromjson(cls, folders, jdict, firstgid=None):
        print('fromjson (isometrically)')
        if 'source' in jdict:
            firstgid = int(jdict['firstgid'])
            srcc = jdict['source']

            # TODO: Another direct disk access here.
            if srcc.endswith(("tsx", "xml")):
                with open(os.path.join(os.pathsep.join(folders), srcc)) as f:
                    print('opened ', srcc)
                    tag = ElementTree.fromstring(f.read())
                    return cls.fromxml(folders, tag, firstgid)
            elif srcc.endswith(("tsj", "json")):
                with open(os.path.join(os.pathsep.join(folders), srcc)) as f:
                    jdict = json.load(f)

        name = jdict['name']
        if firstgid is None:
            firstgid = int(jdict.get('firstgid', 1))
        tile_width = int(jdict['tilewidth'])
        tile_height = int(jdict['tileheight'])
        num_tiles = int(jdict['tilecount'])

        tileset = cls(name, tile_width, tile_height, firstgid)

        if "transformations" in jdict:
            c = jdict["transformations"]
            tileset.vflip = int(c.get("vflip", 0)) == 1
            tileset.hflip = int(c.get("hflip", 0)) == 1

        # TODO: The tileset can only contain an "image" tag or multiple "tile" tags; it can't combine the two.
        # This should be enforced. For now, I'm just gonna support spritesheet tiles.

        # create a tileset
        arg_sheet = jdict['image']
        tileset._add_image(folders, arg_sheet, num_tiles)

        return tileset


class IsometricMapObject:
    """
    A thing that can be placed on the map.
    """

    def __init__(self, **keywords):
        self.name = ""
        self.type = ""
        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.gid = 0
        self.visible = 1
        self.properties = dict()
        super().__init__(**keywords)

    def __call__(self, dest_surface, sx, sy, mymap):
        """Draw this object at the requested surface coordinates on the provided surface."""
        if self.gid:
            tile_id = self.gid & NOT_ALL_FLAGS
            if tile_id > 0:
                my_tile = mymap.tilesets[tile_id]
                my_tile.paint_tile(dest_surface, sx, sy, self.gid & FLIPPED_HORIZONTALLY_FLAG,
                        self.gid & FLIPPED_VERTICALLY_FLAG)

    @staticmethod
    def _deweirdify_coordinates(tx, ty, givenlayer):
        # It took ages to figure out the coordinate system that Tiled uses for objects on isometric maps. At first I
        # thought the pixel coordinate origin would be the upper left corner of the map's bounding box. It isn't.
        # In fact, it isn't a normal cartesian coordinate system at all. The pixel x,y values are the cell index
        # multiplied by the cell height. I cannot think of any situation for which this would be a useful way to store
        # pixel coordinates, but there you go.
        #
        # This function takes the Tiled pixel coordinates and changes them to tilemap cell coordinates. Feel free to
        # delete this long rant of a comment. Or leave it as a warning to others. I am just glad to finally understand
        # what's going on.

        mx = tx / float(givenlayer.tile_height)
        my = ty / float(givenlayer.tile_height)
        return mx, my

    @classmethod
    def fromxml(cls, tag, givenlayer):
        myob = cls()
        myob.name = tag.attrib.get("name")
        myob.type = tag.attrib.get(info_type_obj)
        # Convert the x,y pixel coordinates to x,y map coordinates.
        x = float(tag.attrib.get("x", 0))
        y = float(tag.attrib.get("y", 0))
        myob.x, myob.y = cls._deweirdify_coordinates(x, y, givenlayer)
        myob.gid = int(tag.attrib.get("gid"))
        myob.visible = int(tag.attrib.get("visible", 1))
        for t in tag:
            if t.tag == "properties":
                for p in t:
                    if p.tag == "property":
                        myob.properties[p.attrib.get("name", "property")] = p.attrib.get("value")
        return myob

    @classmethod
    def fromjson(cls, jdict, givenlayer):
        myob = cls()
        myob.name = jdict.get("name")
        myob.type = jdict.get(info_type_obj)
        # Convert the x,y pixel coordinates to x,y map coordinates.
        x = jdict.get("x", 0)
        y = jdict.get("y", 0)
        myob.x, myob.y = cls._deweirdify_coordinates(x, y, givenlayer)
        myob.gid = jdict.get("gid")
        myob.visible = jdict.get("visible")
        if "properties" in jdict:
            for p in jdict["properties"]:
                myob.properties[p.get("name", "property")] = p.get("value")
        return myob


class IsometricLayer:
    flag_csv = False

    def __init__(self, name, visible, map, offsetx=0, offsety=0):
        self.name = name
        self.visible = visible

        self.tile_width = map.tile_width
        self.tile_height = map.tile_height

        self.width = map.width
        self.height = map.height

        self.offsetx = offsetx
        self.offsety = offsety

        self.properties = {}
        self.cells = list()

    def __repr__(self):
        return '<Layer "%s" at 0x%x>' % (self.name, id(self))

    @classmethod
    def emptylayer(cls, name, givenmap):
        layer = cls(
            name, 0, givenmap, 0, 0
        )

        layer.cells = [0, ] * givenmap.height * givenmap.width

        return layer

    @classmethod
    def fromxml(cls, tag, givenmap):
        layer = cls(
            tag.attrib['name'], int(tag.attrib.get('visible', 1)), givenmap,
            int(tag.attrib.get('offsetx', 0)), int(tag.attrib.get('offsety', 0))
        )

        data = tag.find('data')
        if data is None:
            raise ValueError('layer %s does not contain <data>' % layer.name)

        data = data.text.strip()
        data = data.encode()  # Convert to bytes
        # Decode from base 64 and decompress via zlib
        data = decompress(b64decode(data))

        # I ran a test today and there's a slight speed advantage in leaving the cells as a list. It's not a big
        # advantage, but it's just as easy for now to leave the data as it is.
        #
        # I'm changing to a list from a tuple in case destructible terrain or modifiable terrain (such as doors) are
        # wanted in the future.
        layer.cells = list(struct.unpack('<%di' % (len(data) / 4,), data))
        assert len(layer.cells) == layer.width * layer.height

        return layer

    @classmethod
    def fromjson(cls, jdict, givenmap):
        layer = cls(
            jdict['name'], jdict.get('visible', True), givenmap,
            jdict.get('offsetx', 0), jdict.get('offsety', 0)
        )

        data = jdict.get('data')
        if data is None:
            raise ValueError('layer %s does not contain <data>' % layer.name)

        # Remark: (june 2022)
        # we need to support uncompressed data like CSV, for instance,
        # for specific use cases where the zlib module cannot be used
        # (hacking in the web ctx for example)

        if not cls.flag_csv:  # default mode -> data is compressed
            data = data.strip()
            data = data.encode()  # Convert to bytes
            # Decode from base 64 and decompress via zlib
            data = decompress(b64decode(data))
            # I ran a test today and there's a slight speed advantage in leaving the cells as a list.
            # It's not a big advantage, but it's just as easy for now to leave the data as it is.
            # I'm changing to a list from a tuple in case destructible terrain or modifiable terrain
            # (such as doors) ar wanted in the future
            layer.cells = list(struct.unpack('<%di' % (len(data) / 4,), data))
            assert len(layer.cells) == layer.width * layer.height
        else:  # uncompressed data
            layer.cells = data

        return layer

    def __len__(self):
        return self.height * self.width

    def _pos_to_index(self, x, y):
        # loop map feat.
        y = y % self.height
        x = x % self.width
        return y * self.width + x

    def __getitem__(self, key):
        x, y = key
        i = self._pos_to_index(x, y)
        # no test of boundaries bc of the loop map feat.
        # if 0 <= x < self.width and 0 <= y < self.height:
        return self.cells[i]

    def __setitem__(self, pos, value):
        x, y = pos
        i = self._pos_to_index(x, y)
        # if 0 <= x < self.width and 0 <= y < self.height:
        self.cells[i] = value


class ObjectGroup:
    def __init__(self, name, visible, offsetx, offsety):
        self.name = name
        self.visible = visible
        self.offsetx = offsetx
        self.offsety = offsety

        self.contents = list()

    @classmethod
    def fromxml(cls, tag, givenlayer, object_classes=None):
        mygroup = cls(
            tag.attrib['name'], int(tag.attrib.get('visible', 1)),
            int(tag.attrib.get('offsetx', 0)), int(tag.attrib.get('offsety', 0))
        )

        for t in tag:
            if t.tag == "object":
                if object_classes and t.attrib.get(info_type_obj) in object_classes:
                    myclass = object_classes[t.attrib.get(info_type_obj)]
                else:
                    myclass = IsometricMapObject
                mygroup.contents.append(myclass.fromxml(
                    t, givenlayer
                ))

        return mygroup

    @classmethod
    def fromjson(cls, folders, jdict, givenlayer, object_classes=None):
        mygroup = cls(
            jdict.get('name'), jdict.get('visible', True),
            jdict.get('offsetx', 0), jdict.get('offsety', 0)
        )

        if "objects" in jdict:
            for t in jdict["objects"]:
                if object_classes and t.get(info_type_obj) in object_classes:
                    myclass = object_classes[t.get(info_type_obj)]
                else:
                    myclass = IsometricMapObject
                mygroup.contents.append(myclass.fromjson(
                    t, givenlayer
                ))

        return mygroup


class IsometricMap:
    # Customization:
    # If a layer named "Move Layer" exists, the PC can only move into tiles that exist on this layer. The layer may be
    #   invisible.
    # If a layer named "Block Layer" exists, the PC cannot move into tiles that exist on this layer. The layer may be
    #   invisible.
    def __init__(self):
        self.tile_width = 0
        self.tile_height = 0
        self.width = 0
        self.height = 0
        self.layers = list()
        self.tilesets = Tilesets()
        self.objectgroups = dict()

        self.wrap_x = False
        self.wrap_y = False

        self.floor_layer = None
        self.wall_layer = None

        self.wallpaper = None
        self.mapname = None

    def seek_floor_and_wall(self):
        for n, layer in enumerate(self.layers):
            if layer.name == "Move Layer":
                self.floor_layer = n
            elif layer.name == "Block Layer":
                self.wall_layer = n

    @classmethod
    def load_tmx(cls, folders, filename, object_classes=None):
        # object_classes is a function that can parse a dict describing an object.
        # If None, the only objects that can be loaded are terrain objects.
        with open(os.path.join(os.pathsep.join(folders), filename)) as f:
            tminfo_tree = ElementTree.fromstring(f.read())

        # get most general map informations and create a surface
        tilemap = cls()

        tilemap.width = int(tminfo_tree.attrib['width'])
        tilemap.height = int(tminfo_tree.attrib['height'])
        tilemap.tile_width = int(tminfo_tree.attrib['tilewidth'])
        tilemap.tile_height = int(tminfo_tree.attrib['tileheight'])

        for tag in tminfo_tree.findall('tileset'):
            tilemap.tilesets.add(
                IsometricTileset.fromxml(folders, tag)
                # hacks work only if no more than 1 ts
            )

        for tag in tminfo_tree:
            if tag.tag == 'layer':
                layer = IsometricLayer.fromxml(tag, tilemap)
                tilemap.layers.append(layer)
            elif tag.tag == "objectgroup":
                if not tilemap.layers:
                    # If the first layer on the map is an objectgroup, this is gonna be a problem. Without
                    # a frame of reference, we won't be able to know what tile the object is in, and that is going
                    # to be important information. So, we add an empty layer with no offsets to act as this
                    # objectgroup's frame of reference.
                    tilemap.layers.append(IsometricLayer.emptylayer("The Mysterious Empty Layer", tilemap))
                tilemap.objectgroups[tilemap.layers[-1]] = ObjectGroup.fromxml(tag, tilemap.layers[-1], object_classes)
            elif tag.tag == "properties":
                for ptag in tag.findall("property"):
                    if ptag.get("name") == "wrap_x":
                        tilemap.wrap_x = ptag.get("value") == "true"
                    elif ptag.get("name") == "wrap_y":
                        tilemap.wrap_y = ptag.get("value") == "true"
                    elif ptag.get("name") == "mapname":
                        tilemap.mapname = ptag.get("mapname")
                    elif ptag.get("name") == "wallpaper":
                        tilemap.wallpaper = pygame.image.load(os.path.join("assets",ptag.get("value"))).convert_alpha()

        return tilemap

    @classmethod
    def load_json(cls, folders, filename, object_classes=None):
        # object_classes is a function that can parse a dict describing an object.
        # If None, the only objects that can be loaded are terrain objects.

        with open(os.path.join(os.pathsep.join(folders), filename)) as f:
            jdict = json.load(f)
        return cls.from_json_dict(folders, jdict, object_classes)

    @classmethod
    def from_json_dict(cls, folders, jdict, object_classes=None):
        # get most general map informations and create a surface
        tilemap = cls()

        tilemap.width = jdict['width']
        tilemap.height = jdict['height']
        tilemap.tile_width = jdict['tilewidth']
        tilemap.tile_height = jdict['tileheight']

        if "properties" in jdict:
            for tag in jdict["properties"]:
                if tag["name"] == "wrap_x":
                    tilemap.wrap_x = tag.get("value", False)
                elif tag["name"] == "wrap_y":
                    tilemap.wrap_y = tag.get("value", False)
                elif tag["name"] == "mapname":
                    tilemap.mapname = tag.get("value")
                elif tag["name"] == "wallpaper":
                    tilemap.wallpaper = pygame.image.load(os.path.join("assets", tag.get("value"))).convert_alpha()

        for tag in jdict['tilesets']:
            tilemap.tilesets.add(
                IsometricTileset.fromjson(folders, tag)
            )

        for tag in jdict["layers"]:
            if tag["type"] == 'tilelayer':
                layer = IsometricLayer.fromjson(tag, tilemap)
                tilemap.layers.append(layer)
            elif tag["type"] == "objectgroup":
                if not tilemap.layers:
                    # See above comment for why I'm adding an empty layer. TLDR: the objects need a reference frame.
                    tilemap.layers.append(IsometricLayer.emptylayer("The Mysterious Empty Layer", tilemap))
                tilemap.objectgroups[tilemap.layers[-1]] = ObjectGroup.fromjson(
                    folders, tag, tilemap.layers[-1], object_classes
                )
        return tilemap

    @classmethod
    def load(cls, folders, filename, object_classes=None):
        if filename.endswith(("tmx", "xml")):
            mymap = cls.load_tmx(folders, filename, object_classes)
        elif filename.endswith(("tmj", "json")):
            mymap = cls.load_json(folders, filename, object_classes)
        else:
            raise NotImplementedError("No decoder for {}".format(filename))
        mymap.seek_floor_and_wall()
        return mymap

    def on_the_map(self, x, y):
        # Returns true if (x,y) is on the map, false otherwise
        return (self.wrap_x or ((x >= 0) and (x < self.width))) and (self.wrap_y or ((y >= 0) and (y < self.height)))

    def get_layer_by_name(self, layer_name):
        # The Layers type in Kengi supports indexing layers by name, but it
        # doesn't support accessing layers by
        # negative indices. I'm not sure that it supports slicing either. Anyhow, for now, only the map cursor needs
        # to look up layers by name so this function should be good enough for the time being.
        for lay in self.layers:
            if lay.name == layer_name:
                return lay

    def get_object_by_name(self, object_name):
        # Return the first object found with the provided name.
        for obgroup in self.objectgroups.values():
            for ob in obgroup.contents:
                if ob.name == object_name:
                    return ob

    def clamp_pos(self, pos):
        # For infinite scroll maps, clamp the x and/or y values
        nupos = list(pos)
        if self.wrap_x:
            f, i = math.modf(pos[0])
            nupos[0] = int(i) % self.width + f
        else:
            if pos[0] < 0:
                nupos[0] = 0
            elif pos[0] >= self.width:
                nupos[0] = self.width-1
        if self.wrap_y:
            f, i = math.modf(pos[1])
            nupos[1] = int(i) % self.height + f
        else:
            if pos[1] < 0:
                nupos[1] = 0
            elif pos[1] >= self.height:
                nupos[1] = self.height-1
        return tuple(nupos)

    def clamp_pos_int(self, pos):
        # For infinite scroll maps, clamp the x and/or y values
        nupos = [int(math.floor(c)) for c in pos]
        if self.wrap_x:
            f, i = math.modf(pos[0])
            nupos[0] = int(i) % self.width
        else:
            if pos[0] < 0:
                nupos[0] = 0
            elif pos[0] >= self.width:
                nupos[0] = self.width-1
        if self.wrap_y:
            f, i = math.modf(pos[1])
            nupos[1] = int(i) % self.height
        else:
            if pos[1] < 0:
                nupos[1] = 0
            elif pos[1] >= self.height:
                nupos[1] = self.height-1
        return tuple(nupos)

    def tile_is_blocked(self, x, y):
        pos = self.clamp_pos_int((x,y))
        if self.floor_layer is not None and self.layers[self.floor_layer][pos] == 0:
            return True
        if self.wall_layer is not None and self.layers[self.wall_layer][pos] != 0:
            return True
        else:
            return False



import operator
import random
from abc import ABCMeta, abstractmethod


class StandardCard:
    OMEGA_SYM = ('2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K', 'A')
    OMEGA_SUIT = ('c', 'd', 'h', 's')

    def __init__(self, code):
        if (not isinstance(code, str)) or len(code) != 2:
            raise ValueError('please pass a StdCard code to the constructor! Err1')
        sym, suit = code
        if suit not in self.OMEGA_SUIT:
            raise ValueError('StdCard code isnt valid! Err2(suit)')
        if sym not in self.OMEGA_SYM:
            raise ValueError('StdCard code isnt valid! Err3(sym)')
        self._code = code

    @classmethod
    def all_card_codes(cls):
        res = list()
        for sym in cls.OMEGA_SYM:
            for suit in cls.OMEGA_SUIT:
                res.append(sym + suit)
        return res

    @classmethod
    def at_random(cls, excluded_set=None):
        if excluded_set is None:
            excluded_set = set()
        if len(excluded_set) >= 52:
            raise ValueError('draw_card Err: excluded_set is too broad!')
        omega_code0 = set(cls.all_card_codes())
        omega_code = tuple(omega_code0 - excluded_set)
        return cls(random.choice(omega_code))

    @property
    def code(self):
        return self._code

    @property
    def rank(self):
        return self._code[0]

    @property
    def rank_text(self):
        x = self._code[0]

        if x == 'J':
            return 'Jack'
        elif x == 'Q':
            return 'Queen'
        elif x == 'K':
            return 'King'
        elif x == 'A':
            return 'Ace'
        return {
            '2': 'Deuce',
            '3': 'Trey',
            '4': 'Four',
            '5': 'Five',
            '6': 'Six',
            '7': 'Seven',
            '8': 'Eight',
            '9': 'Nine',
            'T': 'Ten'
        }[x]

    @property
    def suit(self):
        return self._code[1]

    @property
    def suit_text(self):
        return {
            'c': 'Clubs',
            'd': 'Diamonds',
            'h': 'Hearts',
            's': 'Spades'
        }[self._code[1]]

    def __str__(self):
        return self.rank_text + ' of ' + self.suit_text


class BaseHandOfCards(metaclass=ABCMeta):
    def __init__(self):
        self.content = []

    @property
    def value(self):
        return self.compute_value()

    @abstractmethod
    def compute_value(self):
        raise NotImplementedError


class PokerHand(BaseHandOfCards):
    """
    A list of 5 cards + its score

    ranked from highest to lowest we have:
    < 1> Royal flush: A, K, Q, J, 10, all the same suit
    < 2> Straight flush: Five cards in a sequence, all in the same suit
    < 3> Four of a kind: All four cards of the same rank
    < 4> Full house: Three of a kind with a pair
    < 5> Flush: Any five cards of the same suit, but not in a sequence
    < 6> Straight: Five cards in a sequence, but not of the same suit
    < 7> Three of a kind: Three cards of the same rank
    < 8> Two pair: Two different pairs
    < 9> One Pair: Two cards of the same rank
    <10> High card: When you haven't made any of the hands above
    """

    def __init__(self, hand):
        super().__init__()
        assert len(hand) == 5
        self.content = hand  # list of cards

    def __str__(self):
        out = ""
        for card in self.content:
            out += str(card) + ", "
        return out

    def __getitem__(self, index):
        return self.content[index]

    def __len__(self):
        return len(self.content)

    def is_straight(self):
        """
        a hand is a straight if, when sorted, the current card's rank + 1 is the same as the next card
        """
        values = list()
        for card in self.content:
            values.append(int(PokerHand.adhoc_mapping(card.rank)))
        values.sort()

        for n in range(0, 4):
            if values[n] + 1 != values[n+1]:
                return False
        return True

    def is_flush(self):
        """a hand is a flush if all the cards are of the same suit
        """
        for suit in StandardCard.OMEGA_SUIT:
            # - debug by tom
            # print(f'test [{suit}]')
            # print(f'  [{self.hand[0].suit}]')
            # print(f'  [{self.hand[1].suit}]')
            # print(f'  [{self.hand[2].suit}]')
            # print(f'  [{self.hand[3].suit}]')
            # print(f'  [{self.hand[4].suit}]')
            if all((
                suit == self.content[0].suit,
                suit == self.content[1].suit,
                suit == self.content[2].suit,
                suit == self.content[3].suit,
                suit == self.content[4].suit
            )):
                return True
        return False

    @staticmethod
    def adhoc_mapping(xx):
        tmp = {
            'T': 10,
            'J': 11,
            'Q': 12,
            'K': 13,
            'A': 14
        }
        if xx in tmp:
            rez = str(tmp[xx])
        else:
            rez = xx.zfill(2)
        return rez

    @property
    def description(self):
        """
        returns an accurate poker term to describe the current poker hand
        """
        score = self.value
        if str(score)[0] == '1':
            return "High Card"
        elif str(score)[0] == '2':
            return "One Pair"
        elif str(score)[0] == '3':
            return "Two Pair"
        elif str(score)[0] == '4':
            return "Three of a Kind"
        elif str(score)[0] == '5':
            return "Straight"
        elif str(score)[0] == '6':
            return "Flush"
        elif str(score)[0] == '7':
            return "Full House"
        elif str(score)[0] == '8':
            return "Four of a Kind"
        elif str(score)[0] == '9':
            return "Straight Flush"

    # redef
    def compute_value(self):
        """
        The first digits represents the type of hand and the rest represent the cards in the hands
        returns an integer that represents a score given to the hand.
        """
        card_count = dict()  # the count of each card rank (we ignore the suit for now)
        for tsym in StandardCard.OMEGA_SYM:
            card_count[tsym] = 0

        for card in self.content:
            x = card.code[0]
            card_count[x] += 1

        # count number of unique cards
        unique_count = 0
        for rankCount in card_count.values():
            if rankCount > 0:
                unique_count += 1

        straight = self.is_straight()
        flush = self.is_flush()
        points = 0

        if straight and flush:
            points = max(points, 9)  # straight flush
        elif flush and not straight:
            points = max(points, 6)  # flush
        elif not flush and straight:
            points = max(points, 5)  # straight

        elif unique_count == 2:
            if max(card_count.values()) == 4:
                points = 8  # four of a kind (2 uniques and 4 are the same)
            elif max(card_count.values()) == 3:
                points = 7  # full house (2 unique and 3 are the same)

        elif unique_count == 3:
            if max(card_count.values()) == 3:
                points = 4  # three of a kind (3 unique and 3 are the same)
            elif max(card_count.values()) == 2:
                points = 3  # two pair (3 uniques and 2 are the same)

        elif unique_count == 4:
            if max(card_count.values()) == 2:
                points = 2  # one pair (4 uniques and 2 are the same)

        elif unique_count == 5:
            points = 1  # high card

        # print out the values of the cards in order from greatest to least with 2 digits for each card
        # in order to generate a point value
        sorted_card_count = sorted(list(card_count.items()), key=operator.itemgetter(1, 0), reverse=True)
        for keyval in sorted_card_count:
            if keyval[1] != 0:
                points = int(str(points) + str(keyval[1] * PokerHand.adhoc_mapping(keyval[0])))
        return points


# ----------------
#  util function that finds the best possible 5-hand poker hand amongst 7 cards
# ----------------
def find_best_ph(possib_cards: list):
    """
    returns an instance of PokerHand
    """
    assert len(possib_cards) == 7
    # There are 21 possible 5 card combinations in a deck of 7
    # for texas hold'em this is a static problem so we can define every possibility below and use the list
    # to build the algorithm
    perms = """\
01234,01235,01236,01245,01246,01256,01345,\
01346,01356,01456,02345,02346,02356,02456,\
03456,12345,12346,12356,12456,13456,23456\
""".split(',')
    best_ph_sofar = None
    best_ph_score = None

    for sigma in perms:
        tmpe = list(sigma)
        indices = list(map(int, tmpe))  # content CONVERTED to ints

        chosen_cards = list()
        for i in indices:
            chosen_cards.append(possib_cards[i])
        ph = PokerHand(chosen_cards)
        if (best_ph_sofar is None) or (best_ph_score < ph.value):
            best_ph_sofar = ph
            best_ph_score = ph.value
    return best_ph_sofar


class BlackjackHand(BaseHandOfCards):
    def __init__(self):
        super().__init__()

    # redefinition
    def compute_value(self):
        """ Checks the value of the cards in the player's or dealer's hand. """
        total_value = 0

        for card in self.content:
            value = card[1:]

            # Jacks, kings and queens are all worth 10, and ace is worth 11
            if value == 'j' or value == 'q' or value == 'k':
                value = 10
            elif value == 'a':
                value = 11
            else:
                value = int(value)

            total_value += value

        if total_value > 21:
            for card in self.content:
                # If the player would bust and he has an ace in his hand, the ace's value is diminished by 10
                # In situations where there are multiple aces in the hand, this checks to see if the total value
                # would still be over 21 if the second ace wasn't changed to a value of one.
                # if it's under 21, there's no need
                # to change the value of the second ace, so the loop breaks.
                if card[1] == 'a':
                    total_value -= 10
                if total_value <= 21:
                    break
        return total_value


class CardDeck:
    """
    By default this should create a deck which contains all 52 cards and returns it
    BUT this should be the most GENERIC class,

    > For example what if I need to model 78-card Tarot Deck,
    if a class for TarotCard is known, this should be easy, etc.

    > Or for example what if I design a new game where 3 jokers are used?
    """

    def __init__(self):
        # by default we create a full deck, it contains every single standard card
        self.content = list()
        self.reset()

        # if contenu_init is None:
        #     contenu = ['sj', 'sq', 'sk', 'sa']  # spades
        #     contenu.extend(['hj', 'hq', 'hk', 'ha'])  # hearts
        #     contenu.extend(['cj', 'cq', 'ck', 'ca'])  # clubs
        #     contenu.extend(['dj', 'dq', 'dk', 'da'])  # diamonds
        #     values = range(2, 11)
        #     for x in values:
        #         spades = "s" + str(x)
        #         hearts = "h" + str(x)
        #         clubs = "c" + str(x)
        #         diamonds = "d" + str(x)
        #         contenu.append(spades)
        #         contenu.append(hearts)
        #         contenu.append(clubs)
        #         contenu.append(diamonds)
        #     self.contenu = contenu
        # else:
        #     self.contenu = contenu_init

    def reset(self):
        del self.content[:]
        for c in StandardCard.all_card_codes():
            self.content.append(StandardCard(c))
        random.shuffle(self.content)

    def deal(self, n=1):
        k = self.size
        if n > k:
            raise ValueError(f'cannot deal {n} cards! Only {k} are available in deck')
        res = list()
        while n > 0:
            res.append(self.content.pop(0))
            n -= 1
        return res

    def __getitem__(self, item):
        return self.content[item]

    @property
    def size(self):
        return len(self.content)

    # anciennement def returnFromDead(self, deadDeck):
    def recois(self, autre_deck):
        """ Appends the cards from the deadDeck to the deck that is in play. This is called when the main deck
        has been emptied. """
        # équivaut à défausser deadDeck dans le paquet courant

        for card in autre_deck.contenu:
            self.contenu.append(card)
        self.shuffle()

        return self.__class__([])


    def get_info_card(self, ind):
        return self.contenu[ind]

    def deck_deal(self, dead_deck):
        """
        Shuffles the deck, takes the top 4 cards off the deck,
        appends them to the player's and dealer's hands, and
        returns the player's and dealer's hands.
        """
        self.shuffle()

        li_dealer_hand, li_player_hand = [], []
        cards_to_deal = 4
        deck = self.contenu
        while cards_to_deal > 0:
            if len(deck) == 0:
                dead_deck = self.recois(dead_deck)

            # deal the first card to the player, second to dealer, 3rd to player, 4th to dealer, based on divisibility
            # (it starts at 4, so it's even first)
            if cards_to_deal % 2 == 0:
                li_player_hand.append(deck[0])
            else:
                li_dealer_hand.append(deck[0])

            del deck[0]
            cards_to_deal -= 1
        cls = self.__class__
        return dead_deck, cls(li_player_hand), cls(li_dealer_hand)

    def hit(self, dealt_deck, hand):
        """ Checks to see if the deck is gone, in which case it takes the cards from
        the dead deck (cards that have been played and discarded)
        and shuffles them in. Then if the player is hitting, it gives
        a card to the player, or if the dealer is hitting, gives one to the dealer."""

        # if the deck is empty, shuffle in the dead deck
        deck = self.contenu
        if len(deck) == 0:
            dealt_deck = self.recois(dealt_deck)

        hand.contenu.append(deck[0])
        del deck[0]
        return dealt_deck, hand


if __name__ == '__main__':
    d = CardDeck()
    print(d.size)
    print(d)

    print(d[0])
    print(d[1])
    print(d[2])
    print('deal 3 cards')
    tmp = d.deal(3)
    print('   ', tmp[0])
    print('   ', tmp[1])
    print('   ', tmp[2])
    print(tmp)
    print(d[0])
    print(d[13])



import json

from .. import _hub
from ..compo import gfx


pygame = _hub.pygame

# - check your .json file to be sure the format is well-specified in the code
# cf it matches what's written below
ENUM_FRAMES_ENTRY = 'seq'
DELAY_ENTRY = 'delay'


class AnimatedSprite(pygame.sprite.Sprite):
    def __init__(self, srcpath):
        super().__init__()
        self.image = None

        self.img_source = srcpath + '.png'
        self.infospath = srcpath + '.json'

        self._twidth, self._theight = None, None
        self._ck_desc = None
        self.total_nb_img = 0

        self._data = None
        self._animations = dict()  # the default anim always has the name "idle"

        self.rect = None

        # for one given animation
        self._curr_anim_name = None
        self._curr_img_list = None
        self._curr_nb_frames = 0
        self.k = 0
        self.delay_per_frame = 100 / 1000  # 100ms by default
        self.stack_time = 0

    def __getattr__(self, name):
        print('warning! no preloading has been done on AnimatedSprite inst. ({})'.format(self.img_source))
        if name == 'image':
            self.preload()
            return self.image

    def preload(self):
        fullsheet = pygame.image.load(self.img_source)
        fs_width, fs_height = fullsheet.get_size()

        with open(self.infospath, 'r') as fptr:
            infos_obj = json.load(fptr)
            self._twidth, self._theight = infos_obj['tilesize']
            self._ck_desc = infos_obj['colorkey']
            padding = int(infos_obj['padding'])

            self.total_nb_img = 12  # TODO is this a bug ?

            # -- algorithm:
            # 1) find how many lines & colums
            # 2) put everything in a temp list
            # 3) filter out empty frames (thx to the given padding value)
            # step one
            ncolumns, nlines = fs_width // self._twidth, fs_height // self._theight
            # step two
            tmp = list()
            colork = pygame.color.Color(self._ck_desc)
            for lrank in range(nlines):
                tmp.extend(
                    gfx.Spritesheet(fullsheet, 1).load_strip(
                        (0, lrank * self._theight, self._twidth, self._theight), ncolumns, colork
                    )
                )
            # step three
            if padding > 0:
                self._data = tmp[:-padding]
            else:
                self._data = tmp
            # done --

            self.total_nb_img = len(self._data)
            print('total_nb_img -- ', self.total_nb_img)
            self._load_anims(infos_obj['animations'])
            self.play('idle')

    def _ensure_list(self, obj):
        if isinstance(obj, str):
            tmp = obj.split('-')
            if len(tmp[1]) == 0:
                a = int(tmp[0])
                b = self.total_nb_img-1
            else:
                a = int(tmp[0])
                b = int(tmp[1])

            return list(range(a, b + 1))
        else:
            return obj

    def _load_anims(self, obj):
        """
        being given a JSON-like structure, example:
        "animations":{
            "idle":{"set":"0-5","delay":100},
            "attack":{"set":[6,7,8,9,10,11],"delay":250}
        }
        this method returns a dict name <> pair a,b
        where a is the list of images,
        b is the interframe delay
        """
        self._animations.clear()

        for k, v in obj.items():
            tmp = list()
            enum_frames = self._ensure_list(v[ENUM_FRAMES_ENTRY])
            for idx in enum_frames:

                if idx >= self.total_nb_img:
                    err_m = 'in animation "{}" given range is {} but, no corresp. data in the SpriteSheet!'.format(
                        k, v[ENUM_FRAMES_ENTRY]
                    )
                    raise ValueError(err_m)
                tmp.append(self._data[idx])
            print('anim {} , img count -- {}'.format(k, len(tmp)))
            self._animations[k] = (tmp, v[DELAY_ENTRY] / 1000)  # defined as millisec

    def play(self, anim_name):
        self._curr_anim_name = anim_name
        self._curr_img_list, self.delay_per_frame = self._animations[anim_name]
        self._curr_nb_frames = len(self._curr_img_list)
        self.image = self._curr_img_list[0]
        if self.rect is None:
            self.rect = self.image.get_rect()

        self.k = 0
        self.stack_time = 0

    def draw(self, screenref):
        # this will bug bc of pygame.transform.scale not properly implemented in pygame_emu (v. 007+)
        screenref.blit(self.image, self.rect.topleft)
        # pygame.draw.rect(screenref, 'red', (self.rect.topleft, (89, 89)))

    def update(self, dt):
        self.stack_time += dt

        if self.stack_time > self.delay_per_frame:
            self.stack_time -= self.delay_per_frame
            self.k += 1
            if self.k >= self._curr_nb_frames:
                self.play('idle')
            self.image = self._curr_img_list[self.k]



import abc

from .util import underscore_format


def enum_builder_generic(to_upper, starting_index, *sequential, **named):
    domaine = range(starting_index, len(sequential) + starting_index)
    enums = dict(zip(sequential, domaine), **named)
    tmp_inv_map = {v: k for k, v in enums.items()}
    tmp_all_codes = domaine

    if to_upper:
        tmp = dict()
        for k, v in enums.items():
            if k == 'inv_map' or k == 'all_codes':
                continue
            tmp[k.upper()] = v
        enums = tmp

    enums['inv_map'] = tmp_inv_map
    enums['all_codes'] = tmp_all_codes
    enums['last_code'] = len(sequential) + starting_index - 1
    return type('Enum', (), enums)


def enum_from_n(n, *sequential, **named):
    return enum_builder_generic(False, n, *sequential, **named)


def enum(*sequential, **named):
    """
    the most used enum builder
    """
    return enum_from_n(0, *sequential, **named)


# --- trick from stackoverflow...
class abstractclassmethod(classmethod):
    __isabstractmethod__ = True

    def __init__(self, callable):
        callable.__isabstractmethod__ = True
        super(abstractclassmethod, self).__init__(callable)


# ------------------------------- matrices -------------------------------
class BaseMatrix(object):
    """
    classe ABSTRAITE
    représentation 1d dune matrice bidimensionnelle de taille connue
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, matr_size, li1d_val_init=None):
        self.width, self.height = matr_size
        nb_elem = self.width * self.height
        if li1d_val_init is not None:
            if len(li1d_val_init) != nb_elem:
                raise ValueError('val pour init matrice incoherentes')
            for elt in li1d_val_init:
                if not self.__class__.isValidValue(elt):
                    raise ValueError('val type doesnt match matrix type, elt={}'.format(elt))
            self.repr_1d = li1d_val_init
        else:
            self.repr_1d = [self.__class__.defaultValue()] * nb_elem

    @abstractclassmethod
    def defaultValue(cls):
        pass

    @abstractclassmethod
    def isValidValue(cls, val):
        pass

    def get_val(self, i, j):
        if self.is_out(i, j):
            raise ValueError('coords {} {} out of bounds for the current matrix'.format(i, j))
        adhoc_ind = j * self.width + i
        return self.repr_1d[adhoc_ind]

    def set_val(self, i, j, val):
        if not self.__class__.isValidValue(val):
            raise ValueError(
                'val {} incompatible avec le type de matrice utilisé ({})'.format(val, self.__class__.__name__)
            )
        ind = j * self.width + i
        self.repr_1d[ind] = val

    def set_all(self, val):
        for i in range(self.width):
            for j in range(self.height):
                self.set_val(i, j, val)

    def is_out(self, i, j):
        return not (0 <= i < self.width and 0 <= j < self.height)

    def get_size(self):
        return self.width, self.height

    def __str__(self):
        res = '{} x {} matrix'.format(self.width, self.height)
        res += "\n"
        for j in range(self.height):
            for i in range(self.width):
                ind = j * self.width + i
                res += '  {}'.format(self.repr_1d[ind])
            res += "\n"
        return res


class BoolMatrix(BaseMatrix):
    # --- redefinition de 2 methodes venant den haut
    @classmethod
    def defaultValue(cls):
        return True

    @classmethod
    def isValidValue(cls, val):
        return isinstance(val, bool)


class IntegerMatrix(BaseMatrix):
    """
    modélise une matrice d'entiers. Par défaut, toutes les cellules sont à 0. On redef 2 methodes venant den haut
    """
    @classmethod
    def defaultValue(cls):
        return 0

    @classmethod
    def isValidValue(cls, val):
        return (val is None) or isinstance(val, int)
# --------------- end of matrices -------------------------


class Stack:
    def __init__(self, given_seq=None):
        if given_seq is None:
            self.fifo_list = []
        else:
            self.fifo_list = list(given_seq)

    def push(self, element):
        self.fifo_list.append(element)
        # - debug
        # print()
        # print('PUSH')
        # print('etat de pile {}'.format(self.fifo_list))

    def top_down_trav(self):
        return reversed(self.fifo_list)

    def bottom_up_trav(self):
        return self.fifo_list

    def pop(self):
        try:
            tmp = self.fifo_list.pop()
            res = tmp
        except IndexError:
            res = None  # the stack's already empty

        return res

    def peek(self):
        try:
            return self.fifo_list[-1]
        except IndexError:
            return None  # the stack's empty

    def count(self):
        return len(self.fifo_list)


# --------------------------------------------
#  generic tree structure
# --------------------------------------------
class Tree:
    def __init__(self, rootnode_ref):
        self._root = rootnode_ref
        self.allnodesinfo = set()  # do not modify this member directly

    @property
    def root(self):
        return self._root

    # TODO
    # @root.setter
    # def root(self, noderef):
    #     self._root = noderef

    def append_value(self, value, parent_node):
        if parent_node is None:
            raise ValueError('[Tree]Error - Cannot use append_value to create a root node!')
        if parent_node not in self.allnodesinfo:
            raise ValueError('[Tree]Error - Trying to append value to a node that doesnt exist!')
        return TreeNode(value, parent_node)

    @property
    def count(self):
        return len(self.allnodesinfo)

    def has_node(self, node_ref):
        return node_ref in self.allnodesinfo

    def node_by_content(self, needle_value):
        """
        returns None if needle_value isnt found
        """
        for node_ref in self.allnodesinfo:
            if needle_value == node_ref.value:
                return node_ref

    def sub_tree(self, node_ref):
        raise NotImplementedError  # TODO, use node cloning

    def __str__(self):
        r = ''
        for node in self._root.traverse():
            d = node.depth
            v = node.value
            spstr = ''.join([' ' for _ in range(d)])
            r += f'{spstr}node[depth:{d}] val={v}' + '\n'
        return r

    # def cut_from_node(self, ref_node):
    #     if ref_node == self.root:
    #         raise ValueError('empty tree not allowed')
    #
    #     if not ref_node.is_leaf():
    #         # if node has childs, we shall cut the whole branch
    #         for c in ref_node.childs:
    #             self.cut_from_node(c)
    #
    #     ref_node.parent.childs.remove(ref_node)
    #     self.__allnodes.remove(ref_node)


class TreeNode:
    def __init__(self, value, parent_ref):
        """
        set parent_ref to None,
        in order to create a root node + the associated tree.

        Important remark: this class doesnt allow for the creation of 'headless' nodes,
        that is nodes that are not bound to a given Tree! It simplifies the implementation
        as it prevents the dev from doing things that make no sense. For instance: trying to
        add an existing node (say, the 1st child of root) to the 2nd child of root
            r
          /  \
         a    b
        doing b.add_child(a) should be always be forbidden,
        so we dont implement an add_child method.
        """
        self.value = value  # data

        if parent_ref:
            self.parent = parent_ref
            self.parent.children.append(self)
            self.tree_ref = parent_ref.tree_ref
        else:
            self.parent = None
            self.tree_ref = Tree(self)
        self.tree_ref.allnodesinfo.add(self)

        self.children = list()  # references to other nodes

    @property
    def depth(self):
        cpt = 0
        r = self
        while not r.is_root():
            r = r.parent
            cpt += 1
        return cpt

    @property
    def child_count(self):
        return len(self.children)

    def is_root(self):
        return self.parent is None

    def is_leaf(self):
        return len(self.children) == 0

    def remove_child(self, child_node):
        # removes parent-child relationship
        if self.tree_ref.has_node(child_node):
            self.children = [child for child in self.children if child is not child_node]
            self.tree_ref.allnodesinfo.remove(child_node)
        else:
            print('***[TreeNode] warning trying to remove a child that does not exist***')

    def traverse(self):
        visited_nodes = list()
        nodes_to_visit = [self]
        while len(nodes_to_visit) > 0:
            curr_node = nodes_to_visit.pop(0)
            visited_nodes.append(curr_node)
            nodes_to_visit += curr_node.children
        return visited_nodes


class StContainer:
    """
    contient toutes les instances de classes qui dérivent de BaseGameState
    """

    def __init__(self):
        self.__setup_done = False
        self.assoc_id_state_obj = dict()

    def setup(self, enum_game_states, stmapping, pymodule=None):
        self.__setup_done = True

        # - initialisation d'après paramètre type: core.enum
        if stmapping:  # but = charger la classe deja identifiee en memoire
            for state_ident, adhoc_cls in stmapping.items():
                print('creating state:', adhoc_cls.__name__)
                obj = adhoc_cls(state_ident)

                # if needed: enable Kata bios state
                if -1 == state_ident:
                    obj.glvars_module = pymodule  # TODO fix architecture,
                    # StContainer shouldnt know details bout SDK feat

                # the regular op
                self.assoc_id_state_obj[state_ident] = obj
        else:

            # old code
            for id_choisi, nom_etat in enum_game_states.inv_map.items():
                class_name = nom_etat + 'State'
                self._auto_find_statecls(id_choisi, nom_etat, class_name)

    def _auto_find_statecls(self, id_choisi, nom_etat, nom_cls):
        pymodule_name = underscore_format(nom_etat)
        pythonpath = 'app.{}.state'.format(pymodule_name)
        print('StContainer is loading a new game state...')
        try:
            pymodule = __import__(pythonpath, fromlist=[nom_cls])
            adhoc_cls = getattr(pymodule, nom_cls)  # class has been retrieved -> ok

            obj = adhoc_cls(id_choisi, nom_etat)
            self.assoc_id_state_obj[id_choisi] = obj

        except ImportError as exc:
            print('ERR: Cannot import State Cls!')
            print()
            print('adhoc module name(conv. to underscore format)={}'.format(pymodule_name))
            print('full path for finding class={}'.format(pythonpath))
            print('target class={}'.format(nom_cls))

    def retrieve(self, identifiant):
        """
        :param identifiant: peut-être aussi bien le code (int) que le nom de classe dédiée (e.g. PlayState)
        :return: instance de BaseGameState
        """

        # construction par nom ou identifiant entier
        # TODO rétablir recherche par nom et non par code...

        # if isinstance(identifiant, str):
        #     gamestate_id = None
        #     for num_id, nom in state_listing.items():
        #         if nom == identifiant:
        #             gamestate_id = num_id
        #             break
        #     if gamestate_id is None:
        #         assert 0, "state name not found: " + identifiant
        # else:
        #     gamestate_id = identifiant
        gamestate_id = identifiant
        return self.assoc_id_state_obj[gamestate_id]



import time
from abc import abstractmethod
from collections import deque as deque_obj

from .compo import vscreen
from . import _hub
from . import struct
from .foundation import defs


enum_builder_generic = struct.enum_builder_generic
gl_unique_manager = None


def create_manager():
    global gl_unique_manager
    if gl_unique_manager is None:
        gl_unique_manager = DeadSimpleManager()
    else:
        print('* warning: second call to event.create_manager()')


def _enum_engine_ev_types(*sequential, **named):
    return enum_builder_generic(True, defs.FIRST_ENGIN_TYPE, *sequential, **named)


EngineEvTypes = _enum_engine_ev_types(
    'LogicUpdate',
    'Paint',
    'RefreshScreen',

    'PushState',  # contient un code state_ident
    'PopState',
    'ChangeState',  # contient un code state_ident

    'GameBegins',  # correspond à l'ancien InitializeEvent
    'GameEnds',  # indique que la sortie de jeu est certaine

    'BtClick',

    'ConvChoice',  # contains value (used in rpgs like niobepolis, conv means conversation)
    'ConvStarts',  # contains convo_obj, portrait
    'ConvEnds',

    'FocusCh',
    'FieldCh',
    'DoAuth',

    'AsyncRecv',  # [num] un N°identification & [msg] un string
    'AsyncSend'  # [num] un N°identification & [msg] un string
)


def enum_ev_types(*sequential, **named):  # Custom events /!\ not engine events
    # this function should be used by the custom game
    return enum_builder_generic(False, defs.FIRST_CUSTO_TYPE, *sequential, **named)


class EventManager:
    @classmethod
    def instance(cls):
        global gl_unique_manager
        return gl_unique_manager


class CgmEvent:
    ETYPE_ATTR_NAME = 'cgm_type'
    ref_enum_custom = None

    def __init__(self, engin_ev_type, **kwargs):
        if engin_ev_type == defs.USEREVENT:
            raise ValueError('ev_type {} is not valid (cgm reserved type)')

        self.type = engin_ev_type
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def inject_custom_names(cls, ref_enum):
        cls.ref_enum_custom = ref_enum

    @staticmethod
    def ext_tev_detection(etype):
        # extended type event detection
        return etype == defs.USEREVENT

    # @classmethod
    # def deserialize(cls, ser):
    #     # print(ser)
    #     tmpli = ser.split('@')
    #     adhoc_type = int(tmpli[0])
    #     dico = json.loads(tmpli[1])
    #     if adhoc_type == EngineEvTypes.PAINT:
    #         dico['screen'] = shared.screen
    #     return cls(adhoc_type, **dico)
    #
    # def serialize(self):
    #     kwargs_cp = dict()
    #     for k, v in self.__dict__.items():
    #         if k == 'type':
    #             pass
    #         elif k == 'screen' and self.type == EngineEvTypes.PAINT:
    #             kwargs_cp['screen'] = None
    #         elif isinstance(v, tuple):
    #             kwargs_cp[k] = list(v)
    #         else:
    #             kwargs_cp[k] = v
    #     return '{}@'.format(self.type) + json.dumps(kwargs_cp)

    # ---------- était utile jadis pour faire passer les event sur la file pygame classique --------
    # TODO decider du sort, now 2022

    # def wrap(self):
    #     """
    #     :return: a corresponding pygame event object, so it can be put on the pygame queue
    #     """
    #     tmp = self.__dict__.copy()
    #     if self.type > USEREVENT:
    #         tmp[self.ETYPE_ATTR_NAME] = self.type
    #         t = USEREVENT
    #     else:
    #         t = self.type
    #     del tmp['type']
    #     return kpygame.PygameEvent(t, tmp)
    #
    # @classmethod
    # def unwrap(cls, pygam_ev_obj):
    #     assert pygam_ev_obj.type == USEREVENT
    #     engin_type = getattr(pygam_ev_obj, cls.ETYPE_ATTR_NAME)
    #
    #     tmp = pygam_ev_obj.dict
    #     del tmp[cls.ETYPE_ATTR_NAME]
    #     return cls(engin_type, **tmp)

    # def __str__(self):
    #     if self.type > defs.USEREVENT:
    #         if self.type < defs.FIRST_CUSTO_TYPE:
    #             nom = EngineEvTypes.inv_map[self.type]
    #         elif self.ref_enum_custom is None:
    #             nom = '????'
    #         else:
    #             nom = self.ref_enum_custom.inv_map[self.type]
    #     else:
    #         nom = 'event classique pygame, code:' + str(self.type)
    #     tmp = self.__dict__.copy()
    #     del tmp['type']
    #     return '<CgmEvent({}-{} {})>'.format(self.type, nom, tmp)


# class CogObject:
#     _instances_clue = set()
#     _unavailable_ids = set()
#     _next_id = 0
#
#     # cached events
#     lu_cached_ev = None
#     paint_cached_ev = None
#
#     # contains time from webctx
#     wt = None
#
#     # we can send 'N times' the logic update event
#     # while keeping the same time.time() value
#     # => this speeds up the game without loosing too much precision on time handling
#     _optim_counter = 0
#     _n_times_same_timeval = 100
#
#     def __init__(self, explicit_id=None):
#         global gl_unique_manager
#
#         # - module events, init procedure...
#         if CogObject.lu_cached_ev is None:
#             CogObject.lu_cached_ev = CgmEvent(EngineEvTypes.LOGICUPDATE, curr_t=None)
#             CogObject.paint_cached_ev = CgmEvent(EngineEvTypes.PAINT, screen=None)
#
#         # resuming constructor
#         self._manager = gl_unique_manager
#
#         # - attribue un nouvel _id pour cette instance de CogObject
#         self._id = self.__trouve_nouvel_id() if (explicit_id is None) else explicit_id
#         self.__class__._unavailable_ids.add(self._id)
#
#         # - garde trace de cette nouvelle instance...
#         self._instances_clue.add(weakref.ref(self))
#
#     @classmethod
#     def __trouve_nouvel_id(cls):
#         while cls._next_id in cls._unavailable_ids:
#             cls._next_id += 1
#         res = cls._next_id
#         return res
#
#     def get_id(self):
#         return self._id
#
#     @classmethod
#     def __list_instances(cls):
#         dead = set()
#         for ref in cls._instances_clue:
#             obj = ref()
#             if obj is not None:
#                 yield obj
#             else:
#                 dead.add(ref)
#         cls._instances_clue -= dead
#
#     @classmethod
#     def select_by_id(cls, given_id):
#         for obj in cls.__list_instances():
#             if given_id == obj.get_id():
#                 return obj
#         print('**warning requested CogObject [id={}] not found'.format(given_id))
#
#     def pev(self, ev_type, **kwargs):
#         if ev_type == EngineEvTypes.LOGICUPDATE:
#
#             if CogObject.wt:
#                 CogObject.lu_cached_ev.curr_t = CogObject.wt
#             else:
#                 CogObject.lu_cached_ev.curr_t = time.time()
#             rdy_ev = CogObject.lu_cached_ev
#
#         elif ev_type == EngineEvTypes.PAINT:
#
#             CogObject.paint_cached_ev.screen = shared.screen
#             rdy_ev = CogObject.paint_cached_ev
#
#         else:
#
#             rdy_ev = CgmEvent(ev_type, **kwargs)
#
#         self._manager.post(rdy_ev)


class CogObj:
    _free_id = 2
    _taken = set()

    def __init__(self, explicit_id=None):
        cls = self.__class__
        if explicit_id:
            if explicit_id in cls._taken:
                raise ValueError('cannot create CogObj with id{}, already taken!'.format(explicit_id))
            self._ident = explicit_id
        else:
            while cls._free_id in cls._taken:
                cls._free_id += 1
            self._ident = cls._free_id
        cls._taken.add(self._ident)

        self._cached_lu = CgmEvent(EngineEvTypes.LOGICUPDATE, curr_t=None)
        self._cached_pt = CgmEvent(EngineEvTypes.PAINT, screen=None)
        self.latest_rank = None

    @classmethod
    def reset_class_state(cls):
        cls._free_id = 2
        cls._taken.clear()

    def get_id(self):
        return self._ident

    def pev(self, ev_type, **kwargs):
        if EngineEvTypes.LOGICUPDATE == ev_type:
            self._cached_lu.curr_t = time.time()
            event_obj = self._cached_lu

        elif EngineEvTypes.PAINT == ev_type:
            c1 = self._cached_pt.screen is None
            c2 = self.latest_rank is not None and self.latest_rank != vscreen.screen_rank
            if c1 or c2:  # need to update the cached screen in the PAINT event
                self.latest_rank = vscreen.screen_rank
                self._cached_pt.screen = vscreen.screen
            event_obj = self._cached_pt
        else:
            event_obj = CgmEvent(ev_type, **kwargs)
        gl_unique_manager.post(event_obj)


class EventReceiver(CogObj):

    def __init__(self, explicit_id=None, sticky=False):
        """
        if sticky==True, the receiver won't be removed when "EventManager.soft_reset()"s
        """
        super().__init__(explicit_id)
        self._manager = gl_unique_manager
        self.sticky = sticky

    @property
    def active(self):
        return self._manager.talks_to(self)

    def turn_on(self):
        self._manager.add_listener(self)
        return self

    def turn_off(self):
        self._manager.remove_listener(self)

    @abstractmethod
    def proc_event(self, ev, source):
        raise NotImplementedError


# class EventDispatcher:
#
#     def __init__(self):
#         self._queue = list()
#
#     def add_listener(self, obj: EventReceiver):
#         self._queue.append(obj)
#
#     def prioritize(self, obj):
#         self._queue.remove(obj)
#         self._queue.insert(0, obj)
#
#     def postpone(self, obj):
#         self._queue.remove(obj)
#         self._queue.append(obj)
#
#     def remove_listener(self, obj):
#         self._queue.remove(obj)
#
#     def soft_reset(self):
#         nl = list()
#         for memb in self._queue:
#             if memb.is_sticky():
#                 nl.append(memb)
#         self._queue = nl
#
#     def hard_reset(self):
#         del self._queue[:]
#
#     def count_listeners(self):
#         return len(self._queue)
#
#     def dispatch(self, cgm_event, dispatcher):
#         """
#         :param cgm_event: instance of CgmEvent class, includes a type
#         :param dispatcher: None denotes the general EventDispatcher
#         """
#         for recv in self._queue:
#             if recv.proc_event(cgm_event, dispatcher):
#                 return


# @Singleton
# class EventManager(EventDispatcher):
#     """
#     the root source for all events
#     """
#
#     def post(self, given_evt):
#         global headless_mode, simu_queue
#         if headless_mode:
#             simu_queue.append(given_evt.wrap())
#         else:
#             pygame.event.post(given_evt.wrap())
#
#     def update(self):
#         global headless_mode, simu_queue
#
#         if headless_mode:  # HEADLESS mode
#             simu_queue.reverse()
#             while len(simu_queue) > 0:
#                 rev = simu_queue.pop()
#                 if not CgmEvent.ext_tev_detection(rev.type):  # ev. classique
#                     self.dispatch(rev, None)
#                 else:
#                     cgm_ev = CgmEvent.unwrap(rev)  # ev. étendu
#                     self.dispatch(cgm_ev, None)
#             return
#
#         rev = pygame.event.poll()
#         while rev.type != pygame.NOEVENT:
#             if not CgmEvent.ext_tev_detection(rev.type):  # ev. classique
#                 self.dispatch(rev, None)
#             else:
#                 cgm_ev = CgmEvent.unwrap(rev)  # ev. étendu
#                 self.dispatch(cgm_ev, None)
#
#             rev = pygame.event.poll()


class ListenerSet:
    def __init__(self):
        self._corresp = dict()
        self._listener_ids = list()

    def add_listener(self, cog_obj):
        key = cog_obj.get_id()
        self._corresp[key] = cog_obj
        self._listener_ids.append(key)

    def remove_listener(self, cog_obj):
        key = cog_obj.get_id()
        self._listener_ids.remove(key)
        del self._corresp[key]

    def test_contains_id(self, k):
        return k in self._listener_ids

    def hard_reset(self):
        del self._listener_ids[:]
        self._corresp.clear()

    def soft_reset(self):
        ids_tbr = set()
        for l_id in self._listener_ids:
            if not self._corresp[l_id].sticky:
                ids_tbr.add(l_id)
        for adhoc_id in ids_tbr:
            self._listener_ids.remove(adhoc_id)
            del self._corresp[adhoc_id]

    def __getitem__(self, k):
        return self._corresp[k]

    def __iter__(self):
        return self._listener_ids.__iter__()


class DeadSimpleManager:
    nb_inst = 0

    def __init__(self):
        self.__class__.nb_inst += 1
        if self.__class__.nb_inst > 1:
            raise ValueError('using 2+ instances of DeadSimpleManager is forbidden')

        self._ref_ls = ListenerSet()
        self.add_listener = self._ref_ls.add_listener
        self.remove_listener = self._ref_ls.remove_listener
        self.hard_reset = self._ref_ls.hard_reset
        self.soft_reset = self._ref_ls.soft_reset

        self.pyg = _hub.pygame

        self._omega_mouse_events = {
            self.pyg.MOUSEMOTION,
            self.pyg.MOUSEBUTTONDOWN,
            self.pyg.MOUSEBUTTONUP
        }
        self._ev_queue = deque_obj()

    def talks_to(self, cogobj):
        return self._ref_ls.test_contains_id(cogobj.get_id())

    def get_pressed_keys(self):
        return _hub.pygame.key.get_pressed()

    def post(self, ev):
        self._ev_queue.appendleft(ev)

    def update(self):
        for alien_ev in _hub.pygame.event.get():
            self._ev_queue.appendleft(alien_ev)

        n = len(self._ev_queue)
        for _ in range(n):
            curr_ev = self._ev_queue.pop()
            for cogobj_id in self._ref_ls:
                blocking = self._ref_ls[cogobj_id].proc_event(curr_ev, None)
                if blocking:
                    continue


# ------ avant ct runners.py
class GameTicker(EventReceiver):
    def __init__(self, xmaxfps=None):
        super().__init__(explicit_id=1)

        self._running = True
        self.pyg = _hub.pygame
        self._clock = self.pyg.time.Clock()
        self.maxfps = xmaxfps

    @property
    def running(self):
        return self._running

    def proc_event(self, ev, source):
        if ev.type == self.pyg.QUIT or ev.type == EngineEvTypes.GAMEENDS:
            self.halt()

    def halt(self):
        self._running = False

    @property
    def clock(self):
        return self._clock

    def loop(self):
        while self._running:  # many iterations with only this line
            self.pev(EngineEvTypes.LOGICUPDATE)
            self.pev(EngineEvTypes.PAINT)

            self._manager.update()
            vscreen.flip()
            self._clock.tick(self.maxfps)


# -  --  - - --
#  basé sur (GameTicker.
# - - -  - -
class StackBasedGameCtrl(EventReceiver):

    def __init__(self, existing_ticker, gamestates_enum, glvars_pymodule, stmapping):
        super().__init__(sticky=True)

        self.ticker = existing_ticker
        # lets build up all gamestates objects
        self._st_container = struct.StContainer()

        # relation avec stcontainer
        self._st_container.setup(gamestates_enum, stmapping, glvars_pymodule)

        if -1 in stmapping:  # TODO fix architecture, engine shouldnt know bout SDK feat
            self.first_state_id = -1
        else:
            self.first_state_id = 0

        self.__state_stack = struct.Stack()

    # redefinition
    def halt(self):
        while self.get_curr_state_ident() is not None:
            self._pop_state()
        self.ticker.halt()

    def get_curr_state_ident(self):
        return self.__state_stack.peek()

    def proc_event(self, ev, source):
        if ev.type == _hub.pygame.QUIT or ev.type == EngineEvTypes.GAMEENDS:
            self.halt()

        elif ev.type == EngineEvTypes.PUSHSTATE:
            state_obj = self._st_container.retrieve(ev.state_ident)
            self._push_state(state_obj)

        elif ev.type == EngineEvTypes.POPSTATE:
            self._pop_state()

        elif ev.type == EngineEvTypes.CHANGESTATE:
            state_obj = self._st_container.retrieve(ev.state_ident)
            self._change_state(state_obj)

    def init_state0(self):
        self.__state_stack.push(self.first_state_id)
        self._st_container.retrieve(self.first_state_id).enter()

    def loop(self):
        self.init_state0()
        print('stack based ctrl loop STARTS!')
        self.ticker.loop()

    # --- ---
    #  MÉTIER
    # --- ---
    def _push_state(self, state_obj):
        tmp = self.__state_stack.peek()
        curr_state = self._st_container.retrieve(tmp)
        curr_state.pause()

        self.__state_stack.push(state_obj.get_id())
        state_obj.enter()

    def _pop_state(self):
        self.__only_the_pop_part()

        # FOLLOW - UP
        if self.__state_stack.count() == 0:
            self.ticker.halt()
            print('state count 0')
        else:
            tmp = self.__state_stack.peek()
            state_obj = self._st_container.retrieve(tmp)
            state_obj.resume()

    def _change_state(self, state_obj):
        self.__only_the_pop_part()

        # FOLLOW - UP
        self.__state_stack.push(state_obj.get_id())
        state_obj.enter()

    # Warning! never, ever call this method without some kind of follow-up (private method)
    def __only_the_pop_part(self):
        tmp = self.__state_stack.pop()
        state_obj = self._st_container.retrieve(tmp)
        state_obj.release()



from .NorrecBrain import NorrecBrain


def run_test():
    npcs2class = {
        'norrec': NorrecBrain
    }
    curr_npc_name = 'norrec'
    pnj_courant = npcs2class[curr_npc_name]()  # build npc instance
    CODE_EXIT = 'Q'

    # print tutorial
    print("simulateur de dialogue, à tout moment vous pouvez saisir '" + CODE_EXIT + "' pour quitter")
    print("le dialogue commence...")
    print('\n' + '-'*44)

    # loop
    SYM_PROMPT = '>>> '
    pnj_courant.say_something()
    saisie = input(SYM_PROMPT)
    while saisie != CODE_EXIT:
        pnj_courant.entree_info(saisie)
        pnj_courant.say_something()
        saisie = input(SYM_PROMPT)
    print('done!')
